# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
CLI to plot the extract and summarise dipolar couplings.

TODO
* implement symmetry/label-based averaging.
* implement averaging over functional groups.
* implement rotational averaging. 
"""

__author__ = "J. Kane Shenton"
__maintainer__ = "J. Kane Shenton"
__email__ = "kane.shenton@stfc.ac.uk"
__date__ = "August 10, 2022"


import logging

import click
import click_log
import numpy as np
import pandas as pd
from ase import io

from soprano.data.nmr import _get_isotope_list
from soprano.properties.labeling import MagresViewLabels
from soprano.properties.nmr import *
from soprano.scripts.cli_utils import (
    DIPOLAR_OPTIONS,
    NO_CIF_LABEL_WARNING,
    add_options,
    apply_df_filtering,
    expand_aliases,
    print_results,
    sortdf,
    units_rename,
    viewimages,
)
from soprano.selection import AtomSelection
from soprano.utils import has_cif_labels

HEADER = """
##################################################
#  Extracting Dipolar couplingsfrom magres file  #
"""
FOOTER = """
#  End of dipolar coupling extraction            #
##################################################

"""
dipolar_aliases = {
    "minimal": [
            "pair",
            "label_i",
            "label_j",
            "isotope_i",
            "isotope_j",
            "D",
            "alpha",
            "beta",
            ],
    "essential": [
        "pair",
        "D"
    ]}
rss_aliases = {
    "minimal": [
            "index",
            "label",
            "isotope",
            "D_RSS",
            ],
     "essential": [
            "index",
            "D_RSS",
            ],}




# logging
logging.captureWarnings(True)
logger = logging.getLogger("cli")
click_log.basic_config(logger)


@click.command()

# one of more magres files
@click.argument("files", nargs=-1, type=click.Path(exists=True), required=True)
@add_options(DIPOLAR_OPTIONS)
def dipolar(
    files,
    selection_i=None,
    selection_j=None,
    rss_flag=False,
    rss_cutoff=5.0,
    self_coupling=False,
    isonuclear=False,
    output=None,
    output_format=None,
    merge=False,
    isotopes={},
    precision=3,
    sortby=None,
    sort_order="ascending",
    include=None,
    exclude=None,
    query=None,
    view=False,
    verbosity=0,
    **kwargs,
):
    """
    Extract and summarise dipolar couplings from structure files.

    Usage:
    soprano dipolar seedname.{magres|cif|POSCAR|etc}
    """
    if verbosity == 0:
        logging.basicConfig(level=logging.WARNING)
    elif verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.DEBUG)

    # set pandas print precision
    pd.set_option("display.precision", precision)
    # make sure we output all rows, even if there are lots!
    pd.set_option("display.max_rows", None)

    dfs = []
    images = []
    # loop over files
    for fname in files:
        logger.info(HEADER)
        logger.info(fname)

        # try to read in the file:
        try:
            atoms = io.read(fname)
        except OSError:
            logger.error(f"Could not read file {fname}, skipping.")
            continue

        # Inform user of best practice RE CIF labels
        if not has_cif_labels(atoms):
            logger.info(NO_CIF_LABEL_WARNING)

        # Selections -- if None, does all combinations
        sel_i = AtomSelection.all(atoms)
        sel_j = AtomSelection.all(atoms)

        # select subset of atoms based on selection string
        if selection_i:
            logger.info(f"\nSelecting atoms based on selection string: {selection_i}")
            sel_i = AtomSelection.from_selection_string(atoms, selection_i)
        if selection_j:
            if rss_flag:
                raise ValueError("Cannot use --rss flag with selection_j")
            logger.info(f"\nSelecting atoms based on selection string: {selection_j}")
            sel_j = AtomSelection.from_selection_string(atoms, selection_j)

        # --- rss or pairs --- #
        if rss_flag:
            df = extract_dipolar_RSS_couplings(
                atoms,
                isotopes=isotopes,
                cutoff=rss_cutoff,
                isonuclear=isonuclear,
                sel_i=sel_i,
            )
            essential_columns = [
                "index",
                "label",
                "isotope",
                "D_RSS",
            ]
            df = apply_df_filtering(
                df,
                expand_aliases(include, rss_aliases),
                exclude,
                query,
                essential_columns=rss_aliases["essential"],
                logger=logger,
            )
        else:
            df = extract_dipolar_couplings(
                atoms,
                sel_i=sel_i,
                sel_j=sel_j,
                isotopes=isotopes,
                rss_cutoff=rss_cutoff,
                isonuclear=isonuclear,
            )

            # reformat the 'v' column
            df["v"] = df["v"].apply(lambda x: np.round(x, precision))


            df = apply_df_filtering(
                df,
                expand_aliases(include, dipolar_aliases),
                exclude,
                query,
                essential_columns=dipolar_aliases["essential"],
                logger=logger,
            )

        ###############################################

        # add file info
        df["file"] = fname

        if len(df) > 0:
            # done -- save to lists
            dfs.append(df)
            images.append(atoms)
            logger.info(FOOTER)

    if view:
        # TODO: make sure indices match the original/df indices (currently don't if organic!)
        viewimages(images)

    if merge:
        # merge all dataframes into one
        dfs = [pd.concat(dfs, axis=0)]
    for i, df in enumerate(dfs):
        dfs[i] = sortdf(df, sortby, sort_order)

    # rename columns to include units for those that have units
    for df in dfs:
        df.rename(columns=units_rename, inplace=True)

    # write to file(s)
    print_results(dfs, output, output_format, verbosity > 0)


def extract_dipolar_RSS_couplings(
    atoms, isotopes={}, cutoff=5.0, isonuclear=False, sel_i=None, **kwargs
):
    """
    Extracts the dipolar couplings from the atoms object.

    We compute the dipolar constant including all atoms within a cutoff radius.
    We only then filter out the atoms that are not in the selection.

    If you instead want to compute the dipolar constant only including isonuclear atoms,
    you can set isonuclear=True.

    :param atoms : ASE atoms object
    :return: a pandas dataframe with the couplings
    """

    # subset to include in the df:
    idx = np.arange(len(atoms))
    if sel_i:
        idx = sel_i.indices

    elements = atoms.get_chemical_symbols()
    isotopelist = _get_isotope_list(elements, isotopes=isotopes, use_q_isotopes=False)
    species = [f"{iso}{el}" for el, iso in zip(elements, isotopelist)]
    dip_RSS = DipolarRSS.get(
        atoms, isotopes=isotopes, cutoff=cutoff, isonuclear=isonuclear
    )
    if atoms.has("labels"):
        labels = atoms.get_array("labels")
    else:
        labels = MagresViewLabels.get(atoms, store_array=True)

    df = pd.DataFrame(
        {
            "index": [atom.index for atom in atoms],
            "label": labels,
            "isotope": species,
            "D_RSS": dip_RSS,
        }
    )

    # filter out the atoms that are not in the selection
    df = df[df["index"].isin(idx)]

    return df


def extract_dipolar_couplings(
    atoms,
    isotopes={},
    self_coupling=False,
    cutoff=5.0,
    isonuclear=False,
    sel_i=None,
    sel_j=None,
    **kwargs,
):
    """
    Extracts the dipolar couplings from the atoms object.
    :param atoms : ASE atoms object
    :return: a pandas dataframe with the couplings
    """

    elements = atoms.get_chemical_symbols()
    isotopelist = _get_isotope_list(elements, isotopes=isotopes, use_q_isotopes=False)
    species = [f"{iso}{el}" for el, iso in zip(elements, isotopelist)]

    dip = DipolarCoupling.get(
        atoms,
        isotopes=isotopes,
        sel_i=sel_i,
        sel_j=sel_j,
        self_coupling=self_coupling,
        isonuclear=isonuclear,
    )
    # if dip is an empty dict, return an empty dataframe
    if len(dip) == 0:
        return pd.DataFrame()

    if atoms.has("labels"):
        labels = atoms.get_array("labels")
    else:
        labels = MagresViewLabels.get(atoms, store_array=True)

    # transform dip into columns
    pairs, values = zip(*sorted(dip.items()))
    idx_i, idx_j = zip(*pairs)
    d, v = zip(*values)
    # convert to np array
    d = np.array(d)
    # d = d * 2 * np.pi # convert to rad/s? <- not conventional in SSNMR!
    d = d * 1e-3  # convert to kHz
    dip_df = pd.DataFrame(
        {
            "pair": pairs,
            "label_i": [labels[i] for i in idx_i],
            "label_j": [labels[j] for j in idx_j],
            "isotope_i": [species[i] for i in idx_i],
            "isotope_j": [species[j] for j in idx_j],
            "D": d,
            "v": v,
        }
    )

    # get angles in degrees
    dip_df["alpha"] = dip_df.v.apply(lambda x: np.arctan2(-x[1], -x[0]) * 180 / np.pi)
    dip_df["beta"] = dip_df.v.apply(lambda x: np.arccos(-x[2]) * 180 / np.pi)
    # wrap back to -180 to 180
    dip_df["alpha"] = dip_df["alpha"].apply(lambda x: (x + 180) % 360 - 180)
    dip_df["beta"] = dip_df["beta"].apply(lambda x: (x + 180) % 360 - 180)
    return dip_df



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of CastepSubmitter class.

A basic "rolling" submitter for Castep calculations, grabbing from one folder
and depositing results in another.
"""


import glob
import os
import shutil
import subprocess as sp
import sys
import tempfile

from ase.calculators.castep import create_castep_keywords

from soprano.hpc.submitter.submit import Submitter
from soprano.utils import safe_communicate, seedname


class CastepSubmitter(Submitter):
    def set_parameters(
        self,
        folder_in,
        folder_out,
        castep_command,
        castep_path=None,
        copy_extensions=[".castep"],
        pspot_files=[],
        dryrun_test=False,
    ):
        """Set the parameters of the CASTEP Submitter

        | Args:
        |   folder_in (str): path of the folder to extract cell files from
        |   folder_out (str): path of the folder where the results will be
        |                     saved
        |   castep_command (str): command used to call the CASTEP executable
        |                         on this system
        |   castep_path (Optional[str]): folder where the CASTEP executable is
        |                                located (if not part of the system
        |                                PATH)
        |   pspot_files (Optional[list[str]]): additional pseudopotential
        |                                      files to be copied in the input
        |                                      temporary folders
        |   copy_extensions (Optional[list[str]]): extensions of output files
        |                                          to copy to the output
        |                                          folder (by default only
        |                                          .castep file)
        |   dryrun_test (Optional[bool]): run a dryrun test on files before
        |                                 actually running the calculation.
        |                                 Off by default.

        """

        # Check for existence of the folders, if not present create them
        try:
            os.mkdir(folder_in)
        except OSError:
            pass

        try:
            os.mkdir(folder_out)
        except OSError:
            pass

        self.folder_in = folder_in
        self.folder_out = folder_out
        self.castep_command = castep_command
        self.pspots = [os.path.abspath(pp) for pp in pspot_files]
        self.cp_ext = copy_extensions
        self.drun = dryrun_test

        self.castpath = castep_path

        # Finally add the castep_path to PATH if needed
        if castep_path is not None:
            sys.path.append(castep_path)

    def start_run(self):
        # Initialize the CASTEP keywords file in a dedicated temporary folder
        self.kwdir = tempfile.mkdtemp()
        self.log("Creating CASTEP keywords in folder " f"{self.kwdir}\n")
        # Avoid the annoying print out to screen!
        _stdout, sys.stdout = sys.stdout, self._log
        create_castep_keywords(
            self.castep_command, os.path.join(self.kwdir, "castep_keywords.py")
        )
        sys.stdout = _stdout
        sys.path.append(self.kwdir)
        self.log("CASTEP keywords created\n")

    def next_job(self):
        """Grab the next job from folder_in"""

        cfile_list = glob.glob(os.path.join(self.folder_in, "*.cell"))
        if len(cfile_list) == 0:
            return None

        cfile = cfile_list[0]
        name = seedname(cfile)
        self.log(f"Starting job {name}\n")
        files = [cfile]
        # Check if .param file is available too
        if os.path.isfile(os.path.join(self.folder_in, name + ".param")):
            files += [os.path.join(self.folder_in, name + ".param")]
        job = {"name": name, "args": {"files": files}}

        return job

    def setup_job(self, name, args, folder):
        """Copy files to temporary folder to prepare for execution"""

        for f in args["files"]:
            shutil.move(f, folder)
        # Also copy the pseudopotentials
        for pp in self.pspots:
            shutil.copy(pp, folder)

        success = True

        self.log(f"Copying files for job {name}\n")

        # Perform dryrun test if required
        if self.drun:
            self.log("Performing DRYRUN\n")
            dry_proc = sp.Popen(
                [self.castep_command, name, "--dryrun"],
                cwd=folder,
                stdout=sp.PIPE,
                stderr=sp.PIPE,
            )
            stdout, stderr = safe_communicate(dry_proc)
            # When it's finished...
            try:
                castfile = open(os.path.join(folder, name + ".castep"))
            except OSError:
                return False
            # Does the file contain the required lines?
            drline1 = "|       DRYRUN finished ...                       |"
            drline2 = "|       No problems found with input files.       |"

            castlines = castfile.readlines()
            success = False
            for i, l in enumerate(castlines):
                if drline1 in l:
                    if drline2 in castlines[i + 1]:
                        success = True
                        break

            # Ok, now remove the CASTEP file
            castfile.close()
            os.remove(castfile.name)

        return success

    def finish_job(self, name, args, folder):
        """Save required output files to the output folder"""

        for cext in self.cp_ext:
            files = glob.glob(os.path.join(folder, "*" + cext))
            for f in files:
                shutil.move(f, os.path.join(self.folder_out, os.path.basename(f)))

    def finish_run(self):
        """Try removing the temporary keywords directory"""
        try:
            shutil.rmtree(self.kwdir)
        except OSError:
            self.log(
                "Could not delete temporary castep_keywords.py"
                f"directory at {self.kwdir}"
            )
        sys.path.remove(self.kwdir)

        if self.castpath is not None:
            sys.path.remove(self.castpath)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module contains generators meant to produce AtomsCollections based
on different criteria.
"""


from soprano.collection.generate.airss import airssGen
from soprano.collection.generate.defect import additionGen, defectGen, substitutionGen
from soprano.collection.generate.linspace import linspaceGen
from soprano.collection.generate.molneigh import molecularNeighbourhoodGen
from soprano.collection.generate.rattle import rattleGen
from soprano.collection.generate.transform import transformGen



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of AtomsProperty class.

A generic template class that specific Properties will inherit from.
"""


import numpy as np

from soprano.collection import AtomsCollection
from soprano.nmr.tensor import NMRTensor, contains_nmr_tensors
from soprano.selection import AtomSelection


class AtomsProperty:

    default_name = "generic_property"
    default_params = {}

    def __init__(self, name=None, **params):
        """Initialize an AtomsProperty and set its parameters.
        The AtomsProperty instance can then be called with a structure as its
        only argument to get the property with the given parameters.

        | Args:
        |   name (str): a name to give to this specific instance of the
        |               property (will be used to store it as array if
        |               requested)
        |   params: named arguments specific to this type of property

        """

        if name is not None:
            self.name = name
        else:
            self.name = self.default_name
        # Validate the passed parameters
        self.params = dict(self.default_params)
        for p in params:
            if p not in self.params:
                raise ValueError(
                    "Invalid argument passed to"
                    f" '{self.__class__.__name__}'"
                )
            else:
                self.params[p] = params[p]

    @classmethod
    def get(cls, s, store_array=False, selection=None, **kwargs):
        """Extract the given property using the default parameters
        on an Atoms object s

        | Args:
        |   s (ase.Atoms or AtomsCollection): the structure or collection
        |                                     from which to extract the
        |                                     property
        |   store_array (bool): if s is a collection, whether to store the
        |                       resulting data as an array in the collection
        |                       using the default name for this property
        |   selection (str): a selection string to filter the atoms or AtomSelection
        |

        | Returns:
        |   property: the value of the property for the given structure or
        |             a list of values if a collection has been passed
        |

        """

        if isinstance(s, AtomsCollection):
            arr = s.all.map(cls.get, selection=selection, **kwargs)
            if store_array:
                s.set_array(cls.default_name, arr)
            return arr
        else:
            params = dict(cls.default_params)
            params.update(kwargs)

            # Handle selection
            if selection is not None and (params.get("selection") is None):
                # If it's a string, convert to AtomSelection
                if isinstance(selection, str):
                    selection = AtomSelection.from_selection_string(s, selection)
                # Apply the selection to get a subset of atoms
                s_subset = selection.subset(s)
                return cls.extract(s_subset, **params)
            else:
                return cls.extract(s, **params)

    @staticmethod
    def extract(s, **params):
        """Extract the given property with given parameters from an Atoms
        object.

        | Args:
        |   s (ase.Atoms): the structure from which to extract the property
        |   params: named arguments specific to this type of property
        |

        | Returns:
        |   property: the value of the property for the given structure and
        |             parameters
        |

        """

        # Do something specific to get the property of interest
        # Then return the value

    def mean(
        self,
        s: list | AtomsCollection,
        axis: int | None = None,
        weights: np.ndarray | None = None,
        **kwargs
    ) -> dict | list | float | np.ndarray | NMRTensor:
        """
        Compute the mean of the property over a list of structures.

        The default behaviours are:
        - For a list of scalars, compute the mean along the specified axis.
        - For a list of dictionaries, compute the mean for each key across all dictionaries.
        - For a list of NMRTensor objects, compute the mean using the NMRTensor.mean method.
        - For a list of arrays, convert to numpy array and then compute the mean along the specified axis.

        Args:
            s (list[ase.Atoms] | AtomsCollection): The structure or collection
                                                   from which to extract the property.
            axis (int | None): Axis along which the means are computed. If None, compute the mean of scalars.
            weights (np.ndarray | None): An array of weights associated with the values.
                                         If specified, the weighted average will be computed.
                                         Must have the same shape as the property values.
            **kwargs: Additional arguments passed to the property's get method.

        Returns:
            dict | float | np.ndarray | NMRTensor: The mean value of the property for the given structures.

        Raises:
            ValueError: If s is not a collection/list, if property values are None,
                        or if there's an incompatible shape for computing the mean.
            TypeError: If the property values are of a type that cannot be averaged.
        """
        property_values = self.get(s, **kwargs)

        if not property_values:
            raise ValueError("No property values found for the given structures")

        # Apply weights if provided
        if weights is not None:
            if len(weights) != len(property_values):
                raise ValueError(
                    f"Length of weights ({len(weights)}) does not match "
                    f"length of property values ({len(property_values)})"
                )

        # Handle dictionary mean
        if isinstance(property_values[0], dict):
            if weights is not None:
                raise NotImplementedError("Weighted mean for dictionaries is not implemented")
            if axis is not None:
                raise NotImplementedError("Mean along an axis for dictionaries is not implemented")

            mean_property: dict = {}
            for key in property_values[0].keys():
                if not all(key in prop for prop in property_values):
                    raise ValueError(f"Key '{key}' not present in all property dictionaries")

                values = [d.get(key) for d in property_values]
                try:
                    mean_values = np.mean(values, axis=axis)
                except (ValueError, TypeError) as e:
                    raise ValueError(f"Cannot compute mean for key '{key}': {e!s}")
                mean_property[key] = mean_values
            return mean_property

        # Handle list or array mean
        if contains_nmr_tensors(property_values):
            return NMRTensor.mean(property_values, weights=weights, axis=axis)

        try:
            if weights is not None:
                return np.average(property_values, axis=axis, weights=weights)
            return np.mean(property_values, axis=axis)
        except (ValueError, TypeError) as e:
            raise ValueError(f"Cannot compute mean of property values: {e!s}")


    def __call__(self, s, store_array=False, selection=None):
        """Calling the AtomsProperty returns the value of the property as
        extracted with the parameters of this specific instance.

        | Args:
        |   s (ase.Atoms or AtomsCollection): the structure or collection
        |                                     from which to extract the
        |                                     property
        |   store_array (bool): if s is a collection, whether to store the
        |                       resulting data as an array in the collection
        |                       using the given name for this instance
        |   selection (str): a selection string to filter the atoms or AtomSelection
        |

        | Returns:
        |   property: the value of the property for the given structure or
        |             a list of values if a collection has been passed
        |

        """

        if isinstance(s, AtomsCollection):
            arr = s.all.map(self.__call__, selection=selection)
            if store_array:
                s.set_array(self.name, arr)
            return arr
        else:
            return self.extract(s, **self.params)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to NMR shieldings/shifts"""


import warnings

import numpy as np

from soprano.nmr import MagneticShielding
from soprano.nmr.utils import (
    _anisotropy,
    _asymmetry,
    _haeb_sort,
    _skew,
    _span,
    _get_tensor_array,
)
from soprano.properties import AtomsProperty

DEFAULT_MS_TAG = "ms"


def _has_ms_check(f):
    # Decorator to add a check for the magnetic shieldings array
    def decorated_f(s, *args, **kwargs):
        tag = kwargs.get('tag', DEFAULT_MS_TAG)
        if not s.has(tag):
            raise RuntimeError(f"The Atoms object does not have a '{tag}' array.")
        return f(s, *args, **kwargs)

    return decorated_f


def tensor_mean_property(property_name):
    """
    Decorator for creating mean methods that extract a specific property from MagneticShielding objects.
    
    Parameters:
      property_name (str): The name of the property to extract from each MagneticShielding object.
                          Must be a valid attribute or property of MagneticShielding.
    
    Returns:
      decorator: A decorator for mean methods
    """
    def decorator(method):
        def wrapper(self, s, axis=None, weights=None, **kwargs):
            # Get the mean MSTensor
            meanTensors = MSTensor().mean(s, axis=axis, weights=weights, **kwargs)
            # If meanTensors is a list of MagneticShielding objects, extract the specified property
            if isinstance(meanTensors, list) and all(isinstance(T, MagneticShielding) for T in meanTensors):
                # Extract the specified property from each tensor
                return np.array([getattr(T, property_name) for T in meanTensors])
            # If meanTensors is a single MagneticShielding object, extract the specified property
            elif isinstance(meanTensors, MagneticShielding):
                # Extract the specified property from the tensor
                return getattr(meanTensors, property_name)
            # If meanTensors is not a list of MagneticShielding objects, raise an error
            else:
                raise ValueError("meanTensors must be a list of MagneticShielding objects")
        return wrapper
    return decorator


class MSTensor(AtomsProperty):
    """
    MSTensor

    Produces a list of MagneticShielding objects containing the magnetic shielding
    tensors for each atom in the system. 
    Requires the Atoms object to have been loaded from a
    .magres file containing the relevant information.

    Parameters:
      order (str):  Order to use for eigenvalues/eigenvectors. Can
                    be 'i' (ORDER_INCREASING), 'd'
                    (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                    'n' (ORDER_NQR). Default is 'i'.
      tag (str, optional): name of the array containing magnetic shielding tensors.
                    Defaults to 'ms'.

    Returns:
      ms_tensors (list): list of MagneticShielding objects

    """
    default_name = "ms_tensors"
    default_params = {"order": MagneticShielding.ORDER_INCREASING, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, order, tag, **kwargs):
        symbols = s.get_chemical_symbols()
        ms_list = _get_tensor_array(s, tag)

        ref_list = [None] * len(ms_list)
        grad_list = [-1] * len(ms_list)
        if "ref" in kwargs:
            ref = kwargs.pop("ref")
            ref_list = [ref[symbol] if isinstance(ref, dict) else ref for symbol in symbols]
        if "grad" in kwargs:
            grad = kwargs.pop("grad")
            grad_list = [grad[symbol] if isinstance(grad, dict) else grad for symbol in symbols]

        ms_tensors = [MagneticShielding(ms, species=symbol, order=order, reference=ref, gradient=grad)
                        for ms, symbol, ref, grad in zip(ms_list, symbols, ref_list, grad_list)]
        return ms_tensors

class MSDiagonal(AtomsProperty):

    """
    MSDiagonal

    Produces an array containing eigenvalues and eigenvectors for the
    symmetric part of each magnetic shielding tensor in the system. By default
    saves them as part of the Atoms' arrays as well.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   save_array (bool): if True, save the diagonalised tensors in the
    |                      Atoms object as an array. By default True.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_diag (np.ndarray): list of eigenvalues and eigenvectors

    """

    default_name = "ms_diagonal"
    default_params = {"save_array": True}

    @staticmethod
    @_has_ms_check
    def extract(s, save_array, tag):

        ms_tensors = _get_tensor_array(s, tag)
        ms_diag = [np.linalg.eigh((ms + ms.T) / 2.0) for ms in ms_tensors]
        ms_evals, ms_evecs = (np.array(a) for a in zip(*ms_diag))

        if save_array:
            s.set_array(f"{tag}_diagonal" + "_evals", ms_evals)
            # Store also the Haeberlen sorted version
            s.set_array(f"{tag}_diagonal" + "_evals_hsort", _haeb_sort(ms_evals))
            s.set_array(f"{tag}_diagonal" + "_evecs", ms_evecs)

        return np.array([dict(zip(("evals", "evecs"), ms)) for ms in ms_diag])

class MSShielding(AtomsProperty):

    """
    MSShielding

    Produces an array containing the magnetic shielding isotropies in a system
    (ppm).

    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   save_array (bool): if True, save the ms_shielding array in the
    |                      Atoms object as an array. By default True.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_shielding (np.ndarray): list of shieldings

    """

    default_name = "ms_shielding"
    default_params = {"save_array": True, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, save_array, tag) -> np.ndarray:

        ms_tensors = _get_tensor_array(s, tag)
        ms_shielding = np.trace(ms_tensors, axis1=1, axis2=2) / 3.0

        if save_array:
            # Save the isotropic shieldings
            s.set_array(f"{tag}_shielding", ms_shielding)

        return ms_shielding

    @tensor_mean_property('isotropy')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the MSShielding property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          ms_shielding_mean (np.ndarray): The mean of the MSShielding property.
        """


class MSShift(AtomsProperty):

    """
    MSShift

    Produces an array containing the chemical shifts (ppm).
    References must be provided for the chemical shifts to be calculated.
    Optionally, the you can also specify the gradient, m
    
    .. math::
        \\delta = \\sigma_{ref} - m\\sigma
    

    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   ref (list/float/dict): reference frequency per element. Must
    |                          be provided.
    |   gradients float/list/dict: usually around -1. Optional.
    |                              Default: -1 for all elements.
    |   save_array (bool): if True, save the ms_shift array in the
    |                      Atoms object as an array. By default True.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_shift (np.ndarray): list of shifts

    """

    default_name = "ms_shift"
    default_params = {"ref": {}, "grad": -1.0, "save_array": True, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, ref, grad, save_array, tag)-> np.ndarray:
        # make sure we have some references set!
        if not ref:
            raise ValueError("No reference provided for chemical shifts")


        # get shieldings
        ms_shieldings = MSShielding.get(s, tag=tag)

        symbols = np.array(s.get_chemical_symbols())

        # --- REFERENCES --- #
        # array to store the reference for each site
        # defaults to zeros
        references = np.zeros(len(s))

        #-- FLOAT --#
        # if we have a single float, we use it for all sites
        if isinstance(ref, float):
            references[:] = ref
            # if the strucure has more than one type of element, we
            # a single reference is probably not what we want
            # so we raise a warning.
            if len(set(symbols)) > 1:
                warnings.warn('Using the same reference for all elements.\n'
                    'That is probably not what you want.')
        #-- DICT --#
        ## Assuming a format like {'C': 100.0, 'H': 200.0}
        elif isinstance(ref, dict):
            for el, val in ref.items():
                references[symbols == el] = val
            # throw a warning if there are any elements not in the dictionary
            missing_refs = set(symbols) - set(ref.keys())
            if missing_refs:
                warnings.warn(f"Elements {missing_refs} are not in the references dictionary"
                                " and their reference will be set to zero.")
        #-- LIST --#
        ## Assuming a format like [100.0, 100.0, 200.0] with one item per site
        elif isinstance(ref, list):
            if len(ref) != len(s):
                raise ValueError("Reference list must have one item per site/\n"
                "Alternatively, use a dictionary with the element as key and the reference as value")
            references[:] = ref[:]
        else:
            raise ValueError("Reference must be a dictionary element: reference"
                                " or a float or a list of"
                                " floats")

        # --- GRADIENTS --- #
        # default gradient of -1 for each site
        gradients = -1 * np.ones(len(s))

        # Do we have gradients set explicitly?
        if grad:
            ## FLOAT ##
            ## the same gradient for all sites
            if isinstance(grad, float):
                gradients[:] = grad
                if len(set(symbols)) > 1:
                    warnings.warn('You are using the same gradient for all '
                    'elements while referencing the chemical shift.')
            ## DICT ##
            ## Assuming a format like {'C': -1, 'H': -0.98}
            elif isinstance(grad, dict):
                for el, val in grad.items():
                    gradients[symbols == el] = val
            ## LIST ##
            ## Assuming a format like [-1, -1, -0.98] with one item per site
            elif isinstance(grad, list):
                if len(grad) != len(s):
                    raise ValueError("Gradient list must be the same length as the system"
                    "Or use a dictionary with the element as key and the gradient as value")
                gradients[:] = grad[:]
            else:
                raise ValueError("Gradients must be a dictionary element: gradient"
                                " or a float or a list of"
                                " floats")
        # if any of the gradients is outside -1.5 to -0.5 we need to
        # raise a warning
        if np.any(gradients < -1.5) or np.any(gradients > -0.5):
            warnings.warn("Gradients are outside the range: -1.5 to -0.5.\n"
                            "That's a surprising value! Please double check the"
                            "gradients.\n"
                            f"You provided:\n {grad}")


        # Convert from shielding to chemical shift
        ms_shifts = references + (np.array(gradients) * np.array(ms_shieldings)) / (1 + references * 1e-6)

        if save_array:
            # Save the isotropic shifts
            s.set_array(f"{tag}_shift", ms_shifts)


        return ms_shifts

    @tensor_mean_property('shift')
    def mean(self, s, axis=None, weights=None, **kwargs):
        """
        Calculate the mean of the MSShift property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.
          **kwParameters: ref and grad parameters for the MSShift calculation. For example,
                    ref={'C': 100.0, 'H': 200.0} and grad=-1.0.

        Returns:
          ms_shift_mean (np.ndarray): The mean of the MSShift property.
        """


class MSIsotropy(AtomsProperty):

    """
    MSIsotropy

    Produces an array containing the magnetic shielding isotropies in a system
    (ppm).
    If references are provided, the returned values represent the chemical shift.

    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    Refactored into MSShielding and MSShift. This remains here for backwards
    compatibility.

    | Parameters:
    |   ref (float/dict): reference frequency per element. If provided, the chemical shift
    |                will be returned instead of the magnetic shielding.
    |   gradients float/list/dict: usually around -1. 
    |   save_array (bool): if True, save the diagonalised tensors in the
    |                      Atoms object as an array. By default True.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_iso (np.ndarray): list of shieldings/shifts

    """

    default_name = "ms_isotropy"
    default_params = {"ref": {}, "grad": -1.0, "save_array": True, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, ref, grad, save_array, tag) -> np.ndarray:

        if ref:
            # the user wants to use the chemical shift
            ms_iso = MSShift.get(s, ref=ref, grad=grad, save_array=save_array, tag=tag)
        else:
            # the user wants to use the magnetic shielding
            ms_iso = MSShielding.get(s, save_array=save_array, tag=tag)

        if save_array:
            # Save the isotropic shifts
            s.set_array(f"{tag}_isotropy", ms_iso)
        return ms_iso

    def mean(self, s, axis=None, weights=None, **kwargs):
        """
        Calculate the mean of the MSIsotropy property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.
          **kwParameters: ref and grad parameters for the MSIsotropy calculation. For example,
                    ref={'C': 100.0, 'H': 200.0} and grad=-1.0.

        Returns:
          ms_iso_mean (np.ndarray): The mean of the MSIsotropy property.
        """
        # if references are provided in kwargs, we need to calculate the chemical shift
        if kwargs.get("ref"):
            return MSShift().mean(s, axis=axis, weights=weights, **kwargs)
        # otherwise we return the isotropic shielding
        else:
            return MSShielding().mean(s, axis=axis, weights=weights, **kwargs)


class MSAnisotropy(AtomsProperty):

    """
    MSAnisotropy

    Produces an array containing the magnetic shielding anisotropies in a
    system (ppm).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_list (np.ndarray): list of anisotropies

    """

    default_name = "ms_anisotropy"
    default_params = {"force_recalc": False, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc, tag)-> np.ndarray:

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            MSDiagonal.get(s, tag=tag)

        ms_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _anisotropy(ms_evals)

    @tensor_mean_property('anisotropy')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the MSAnisotropy property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          ms_aniso_mean (np.ndarray): The mean of the MSAnisotropy property.
        """
        # Implementation handled by decorator


class MSReducedAnisotropy(AtomsProperty):

    """
    MSReducedAnisotropy

    Produces an array containing the magnetic shielding reduced anisotropies
    in a system (ppm).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_list (np.ndarray): list of reduced anisotropies

    """

    default_name = "ms_red_anisotropy"
    default_params = {"force_recalc": False, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc, tag)-> np.ndarray:

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            MSDiagonal.get(s, tag=tag)

        ms_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _anisotropy(ms_evals, reduced=True)

    @tensor_mean_property('reduced_anisotropy')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the MSReducedAnisotropy property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          ms_red_aniso_mean (np.ndarray): The mean of the MSReducedAnisotropy property.
        """
        # Implementation handled by decorator


class MSAsymmetry(AtomsProperty):

    """
    MSAsymmetry

    Produces an array containing the magnetic shielding asymmetries
    in a system (adimensional).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_list (np.ndarray): list of asymmetries

    """

    default_name = "ms_asymmetry"
    default_params = {"force_recalc": False, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc, tag)-> np.ndarray:

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            MSDiagonal.get(s, tag=tag)

        ms_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _asymmetry(ms_evals)

    @tensor_mean_property('asymmetry')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the MSAsymmetry property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          ms_asym_mean (np.ndarray): The mean of the MSAsymmetry property.
        """
        # Implementation handled by decorator


class MSSpan(AtomsProperty):

    """
    MSSpan

    Produces an array containing the magnetic shielding tensor span
    in a system (ppm).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_list (np.ndarray): list of spans

    """

    default_name = "ms_span"
    default_params = {"force_recalc": False, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc, tag):

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            MSDiagonal.get(s, tag=tag)

        ms_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _span(ms_evals)

    @tensor_mean_property('span')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the MSSpan property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          ms_span_mean (np.ndarray): The mean of the MSSpan property.
        """
        # Implementation handled by decorator


class MSSkew(AtomsProperty):

    """
    MSSkew

    Produces an array containing the magnetic shielding tensor skew
    in a system.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_list (np.ndarray): list of skews

    """

    default_name = "ms_skew"
    default_params = {"force_recalc": False, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc, tag):

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            MSDiagonal.get(s, tag=tag)

        ms_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _skew(ms_evals)

    @tensor_mean_property('skew')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the MSSkew property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          ms_skew_mean (np.ndarray): The mean of the MSSkew property.
        """
        # Implementation handled by decorator


class MSEuler(AtomsProperty):

    """
    MSEuler

    Produces an array of Euler angles in radians expressing the orientation of
    the MS tensors with respect to the cartesian axes for each site in the Atoms object.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.


    Parameters:
        order (str):  Order to use for eigenvalues/eigenvectors. Can
                        be 'i' (ORDER_INCREASING), 'd'
                        (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                        'n' (ORDER_NQR). Default is 'h' for MS tensors.
        convention (str): 'zyz' or 'zxz' accepted - the ordering of the Euler
                        angle rotation axes. Default is ZYZ 
        passive (bool):  active or passive rotations. Default is active (passive=False)
        tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.
         

    Returns:
        ms_eulers (np.array): array of Euler angles in radians

    """

    default_name = "ms_eulers"
    default_params = {"order": MagneticShielding.ORDER_HAEBERLEN,
                      "convention": "zyz",
                      "passive": False,
                      "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, order, convention, passive, tag):
        return np.array([t.euler_angles(convention, passive=passive) for t in MSTensor.get(s, order=order, tag=tag)])

    def mean(self, s, axis=None, weights=None, **kwargs):
        """
        Calculate the mean of the MSEuler property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          ms_euler_mean (np.ndarray): The mean of the MSEuler property.
        """
        # Get the mean MSTensor
        meanTensors = MSTensor().mean(s, axis=axis, weights=weights, **kwargs)

        # If meanTensors is a list of MagneticShielding objects, extract the Euler angles
        if isinstance(meanTensors, list) and all(isinstance(T, MagneticShielding) for T in meanTensors):
            # Extract the Euler angles from each tensor
            return np.array([t.euler_angles(**kwargs) for t in meanTensors])
        # If meanTensors is a single MagneticShielding object, extract the Euler angles
        elif isinstance(meanTensors, MagneticShielding):
            # Extract the Euler angles from the tensor
            return meanTensors.euler_angles(**kwargs)
        # If meanTensors is not a list of MagneticShielding objects, raise an error
        else:
            raise ValueError("meanTensors must be a list of MagneticShielding objects")


class MSQuaternion(AtomsProperty):

    """
    MSQuaternion

    Produces a list of ase.Quaternion objects expressing the orientation of
    the MS tensors with respect to the cartesian axes.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    This is now deprecated in favour of an explicit Euler angle calculation
    that better handles NMR tensors.

    | Parameters:
    |   order (str):  Order to use for eigenvalues/eigenvectors. Can
                        be 'i' (ORDER_INCREASING), 'd'
                        (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                        'n' (ORDER_NQR). Default is 'i'.
    |   tag (str): name of the array containing magnetic shielding tensors. Default: 'ms'.

    | Returns:
    |   ms_quat (list): list of quaternions

    """

    default_name = "ms_quats"
    default_params = {"order": MagneticShielding.ORDER_HAEBERLEN, "tag": DEFAULT_MS_TAG}

    @staticmethod
    @_has_ms_check
    def extract(s, order, tag):
        return [t.quaternion for t in MSTensor.get(s, order=order, tag=tag)]

    @tensor_mean_property('quaternion')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the MSQuaternion property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          ms_quat_mean (np.ndarray): The mean of the MSQuaternion property.
        """
        # Implementation handled by decorator


# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to NMR electric field
gradients"""

from collections import defaultdict
from typing import List

import numpy as np

from soprano.data.nmr import EFG_TO_CHI, _get_isotope_data, _get_isotope_list
from soprano.nmr import ElectricFieldGradient
from soprano.nmr.utils import (
    _anisotropy,
    _asymmetry,
    _frange,
    _haeb_sort,
    _skew,
    _span,
    _get_tensor_array,
)
from soprano.properties import AtomsProperty


DEFAULT_EFG_TAG = "efg"

def _has_efg_check(f):
    # Decorator to add a check for the electric field gradient array
    def decorated_f(s, *args, **kwargs):
        # Extract tag from kwargs, default to DEFAULT_EFG_TAG
        tag = kwargs.get("tag", DEFAULT_EFG_TAG)
        if not s.has(tag):
            raise RuntimeError(f"The Atoms object does not have a '{tag}' array.")
        return f(s, *args, **kwargs)

    return decorated_f


def tensor_mean_property(property_name):
    """
    Decorator for creating mean methods that extract a specific property from ElectricFieldGradient objects.
    
    Parameters:
      property_name (str): The name of the property to extract from each ElectricFieldGradient object.
                          Must be a valid attribute or property of ElectricFieldGradient.
    
    Returns:
      decorator: A decorator for mean methods
    """
    def decorator(method):
        def wrapper(self, s, axis=None, weights=None, **kwargs):
            # Get the mean EFGTensor
            meanTensors = EFGTensor().mean(s, axis=axis, weights=weights, **kwargs)

            # If meanTensors is a list of ElectricFieldGradient objects, extract the specified property
            if isinstance(meanTensors, list) and all(isinstance(T, ElectricFieldGradient) for T in meanTensors):
                # Extract the specified property from each tensor
                return np.array([getattr(T, property_name) for T in meanTensors])
            # If meanTensors is a single ElectricFieldGradient object, extract the specified property
            elif isinstance(meanTensors, ElectricFieldGradient):
                # Extract the specified property from the tensor
                return getattr(meanTensors, property_name)
            # If meanTensors is not a list of ElectricFieldGradient objects, raise an error
            else:
                raise ValueError("meanTensors must be a list of ElectricFieldGradient objects")

            # Extract the specified property from each tensor
            return np.array([getattr(T, property_name) for T in meanTensors])
        return wrapper
    return decorator


class EFGTensor(AtomsProperty):
    """
    EFGTensor

    Produces a list of ElectricFieldGradient objects containing the electric field
    gradient tensors for each atom in the system.
    Requires the Atoms object to have
    been loaded from a .magres file containing the relevant information.

    The default convention for EFG tensors is the NQR one (`|Vxx| ≤ |Vyy| ≤ |Vzz|`).
    This is different from the default convention for MS tensors (Haeberlen).
    You can change this by specifying the 'order' parameter.

    Parameters:
        order (str):  Order to use for eigenvalues/eigenvectors. Can
                    be 'i' (ORDER_INCREASING), 'd'
                    (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                    'n' (ORDER_NQR). Default is 'n'.
        use_q_isotopes (bool): if True, always use the most common quadrupole
                             active isotope for each element, if there is
                             one.
        isotopes (dict): dictionary of specific isotopes to use, by element
                         symbol. If the isotope doesn't exist an error will
                          be raised. Overrides use_q_isotopes.
        isotope_list (list): list of isotopes, atom-by-atom. To be used if
                           different atoms of the same element are supposed
                           to be of different isotopes. Where a 'None' is
                           present will fall back on the previous
                           definitions. Where an isotope is present it
                           overrides everything else.
        tag (str, optional): name of the array containing electric field gradient tensors.
                            Defaults to 'efg'.

    Returns:
      efg_tensors (list): list of ElectricFieldGradient objects

    """

    default_name = "efg_tensors"
    default_params = {
        "order": ElectricFieldGradient.ORDER_NQR,
        "use_q_isotopes": False,
        "isotopes": {},
        "isotope_list": None,
        "tag": DEFAULT_EFG_TAG,
    }

    @staticmethod
    @_has_efg_check
    def extract(s, order, use_q_isotopes, isotopes, isotope_list, tag, **kwargs) -> List[ElectricFieldGradient]:

        # First thing, build the isotope dictionary
        elems = s.get_chemical_symbols()

        # Is isotope list valid?
        if isotope_list is not None and len(isotope_list) != len(elems):
            print("WARNING - invalid isotope_list, ignoring")
            isotope_list = None

        # Get the isotope list given the parameters
        isotopelist = _get_isotope_list(elems, isotopes=isotopes, isotope_list=isotope_list, use_q_isotopes=use_q_isotopes)
        # convert list of numbers to isotope symbols
        isotopelist = [f"{iso}{elem}" for iso, elem in zip(isotopelist, elems)]

        efg_tensors_array = _get_tensor_array(s, tag)
        efg_tensors = [ElectricFieldGradient(efg, species, order=order) for efg, species in zip(efg_tensors_array, isotopelist)]
        return efg_tensors


class EFGDiagonal(AtomsProperty):

    """
    EFGDiagonal

    Produces an array containing eigenvalues and eigenvectors for the
    symmetric part of each EFG tensor in the system. By default
    saves them as part of the Atoms' arrays as well.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    Parameters:
      save_array (bool): if True, save the diagonalised tensors in the
                         Atoms object as an array. By default True.
      tag (str): name of the array containing electric field gradient tensors. Default: 'efg'.

    Returns:
      efg_diag (np.ndarray): list of eigenvalues and eigenvectors

    """

    default_name = "efg_diagonal"
    default_params = {
        "save_array": True,
        "tag": DEFAULT_EFG_TAG,
    }

    @staticmethod
    @_has_efg_check
    def extract(s, save_array, tag):

        efg_tensors = _get_tensor_array(s, tag)
        efg_diag = [np.linalg.eigh((efg + efg.T) / 2.0) for efg in efg_tensors]
        efg_evals, efg_evecs = (np.array(a) for a in zip(*efg_diag))

        if save_array:
            s.set_array(f"{tag}_diagonal" + "_evals", efg_evals)
            # Store also the Haeberlen sorted version
            s.set_array(
                f"{tag}_diagonal" + "_evals_hsort", _haeb_sort(efg_evals)
            )
            s.set_array(f"{tag}_diagonal" + "_evecs", efg_evecs)

        return np.array([dict(zip(("evals", "evecs"), efg)) for efg in efg_diag])

    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the EFGDiagonal property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          efg_diag_mean (np.ndarray): The mean of the EFGDiagonal property.
        """
        raise NotImplementedError("Mean calculation for EFGDiagonal is not yet implemented.")


class EFGVzz(AtomsProperty):

    """
    EFGVzz

    Produces an array containing the major component of the electric field
    gradient in a system (au).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    Parameters:
      force_recalc (bool): if True, always diagonalise the tensors even if
                           already present.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      efg_list (np.ndarray): list of Vzz values
    """

    default_name = "efg_vzz"
    default_params = {"force_recalc": False, "tag": DEFAULT_EFG_TAG}

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, tag):

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            EFGDiagonal.get(s, tag=tag)

        efg_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")
        return efg_evals[:, -1]

    @tensor_mean_property('Vzz')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the EFGVzz property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          efg_vzz_mean (np.ndarray): The mean of the EFGVzz property.
        """
        # Implementation handled by decorator


class EFGAnisotropy(AtomsProperty):

    """
    EFGAnisotropy

    Produces an array containing the electric field gradient anisotropies in a
    system (au).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    Parameters:
      force_recalc (bool): if True, always diagonalise the tensors even if
                           already present.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      efg_list (np.ndarray): list of anisotropies
    """

    default_name = "efg_anisotropy"
    default_params = {"force_recalc": False, "tag": DEFAULT_EFG_TAG}

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, tag):

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            EFGDiagonal.get(s, tag=tag)

        efg_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _anisotropy(efg_evals)

    @tensor_mean_property('anisotropy')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the EFGAnisotropy property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          efg_aniso_mean (np.ndarray): The mean of the EFGAnisotropy property.
        """
        # Implementation handled by decorator


class EFGReducedAnisotropy(AtomsProperty):

    """
    EFGReducedAnisotropy

    Produces an array containing the electric field gradient reduced
    anisotropies in a system (au).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    Parameters:
      force_recalc (bool): if True, always diagonalise the tensors even if
                           already present.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      efg_list (np.ndarray): list of reduced anisotropies
    """

    default_name = "efg_red_anisotropy"
    default_params = {"force_recalc": False, "tag": DEFAULT_EFG_TAG}

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, tag):

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            EFGDiagonal.get(s, tag=tag)

        efg_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _anisotropy(efg_evals, reduced=True)

    @tensor_mean_property('reduced_anisotropy')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the EFGReducedAnisotropy property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          efg_red_aniso_mean (np.ndarray): The mean of the EFGReducedAnisotropy property.
        """
        # Implementation handled by decorator


class EFGAsymmetry(AtomsProperty):

    """
    EFGAsymmetry

    Produces an array containing the electric field gradient asymmetries
    in a system (adimensional).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    Parameters:
      force_recalc (bool): if True, always diagonalise the tensors even if
                           already present.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      efg_list (np.ndarray): list of asymmetries
    """

    default_name = "efg_asymmetry"
    default_params = {"force_recalc": False, "tag": DEFAULT_EFG_TAG}

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, tag):

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            EFGDiagonal.get(s, tag=tag)

        efg_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _asymmetry(efg_evals)

    @tensor_mean_property('asymmetry')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the EFGAsymmetry property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          efg_asymm_mean (np.ndarray): The mean of the EFGAsymmetry property.
        """
        # Implementation handled by decorator


class EFGSpan(AtomsProperty):

    """
    EFGSpan

    Produces an array containing the electric field gradient tensor span
    in a system (au).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    Parameters:
      force_recalc (bool): if True, always diagonalise the tensors even if
                           already present.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      efg_list (np.ndarray): list of spans
    """

    default_name = "efg_span"
    default_params = {"force_recalc": False, "tag": DEFAULT_EFG_TAG}

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, tag):

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            EFGDiagonal.get(s, tag=tag)

        efg_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _span(efg_evals)

    @tensor_mean_property('span')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the EFGSpan property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          efg_span_mean (np.ndarray): The mean of the EFGSpan property.
        """
        # Implementation handled by decorator


class EFGSkew(AtomsProperty):

    """
    EFGSkew

    Produces an array containing the electric field gradient tensor skew
    in a system.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    Parameters:
      force_recalc (bool): if True, always diagonalise the tensors even if
                           already present.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      efg_list (np.ndarray): list of skews
    """

    default_name = "efg_skew"
    default_params = {"force_recalc": False, "tag": DEFAULT_EFG_TAG}

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, tag):

        if not s.has(f"{tag}_diagonal" + "_evals_hsort") or force_recalc:
            EFGDiagonal.get(s, tag=tag)

        efg_evals = s.get_array(f"{tag}_diagonal" + "_evals_hsort")

        return _skew(efg_evals)

    @tensor_mean_property('skew')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the EFGSkew property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          efg_skew_mean (np.ndarray): The mean of the EFGSkew property.
        """
        # Implementation handled by decorator


class EFGQuadrupolarConstant(AtomsProperty):

    """
    EFGQuadrupolarConstant

    Produces an array containing the quadrupolar constant in Hz for every atom
    in a system. The constant will be zero for non-quadrupole active nuclei.
    Unless specified otherwise, the quadrupole moment of the most common
    NMR-active isotope is used.
    For reference: the value returned by this property is defined as

    .. math::

        \\frac{e^2qQ}{h}

    in Hz. It is important to keep in mind that therefore this represents a
    *frequency*; the corresponding 'omega' (pulsation) would be the same value
    multiplied by 2*pi. This is, for example, exactly the value required as
    input in Simpson's SPINSYS section.

    Parameters:
      force_recalc (bool): if True, always diagonalise the tensors even if
                           already present.
      use_q_isotopes (bool): if True, always use the most common quadrupole
                             active isotope for each element, if there is
                             one.
      isotopes (dict): dictionary of specific isotopes to use, by element
                       symbol. If the isotope doesn't exist an error will
                       be raised. Overrides use_q_isotopes.
      isotope_list (list): list of isotopes, atom-by-atom. To be used if
                           different atoms of the same element are supposed
                           to be of different isotopes. Where a 'None' is
                           present will fall back on the previous
                           definitions. Where an isotope is present it
                           overrides everything else.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      q_list (np.ndarray): list of quadrupole constants in Hz
    """

    default_name = "efg_qconst"
    default_params = {
        "force_recalc": False,
        "use_q_isotopes": False,
        "isotopes": {},
        "isotope_list": None,
        "tag": DEFAULT_EFG_TAG,
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, use_q_isotopes, isotopes, isotope_list, tag):


        # First thing, build the isotope dictionary
        elems = s.get_chemical_symbols()

        # Is isotope list valid?
        if isotope_list is not None and len(isotope_list) != len(elems):
            print("WARNING - invalid isotope_list, ignoring")
            isotope_list = None

        q_list = _get_isotope_data(elems, "Q", isotopes, isotope_list, use_q_isotopes)

        return EFG_TO_CHI * q_list * EFGVzz.get(s, tag=tag, force_recalc=force_recalc)

    @tensor_mean_property('Cq')
    def mean(self, s, axis=None, weights=None, **kwargs):
        """
        Calculate the mean of the EFGQuadrupolarConstant property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.
          **kwargs: Additional parameters, including isotopes and use_q_isotopes.

        Returns:
          efg_qconst_mean (np.ndarray): The mean of the EFGQuadrupolarConstant property.
        """
        # Implementation handled by decorator


class EFGNQR(AtomsProperty):

    """
    EFGNQR

    Produces an array containing NQR transition frequencies (in Hz) for every atom
    in a system. For non-quadrupole active nuclei, the we return an empty dictionary.
    Unless specified otherwise, the spin and quadrupole moment of the most common
    NMR-active isotope is used.

    For reference: the value returned by this property is defined as

    .. math::

        A = \\frac{V_{zz} Q}{4I(2I - 1)}
        fq = 3A(2m+1)\\sqrt{1 + \\eta^2/3}

    in Hz.
    It is important to keep in mind that therefore this represents a
    *frequency*; the corresponding 'omega' (pulsation) would be the same value
    multiplied by 2*pi. 
    TODO: double-check convention
    TODO: better data structure for the output?

    Parameters:
      force_recalc (bool): if True, always diagonalise the tensors even if
                           already present.
      use_q_isotopes (bool): if True, always use the most common quadrupole
                             active isotope for each element, if there is
                             one.
      isotopes (dict): dictionary of specific isotopes to use, by element
                       symbol. If the isotope doesn't exist an error will
                       be raised. Overrides use_q_isotopes.
      isotope_list (list): list of isotopes, atom-by-atom. To be used if
                           different atoms of the same element are supposed
                           to be of different isotopes. Where a 'None' is
                           present will fall back on the previous
                           definitions. Where an isotope is present it
                           overrides everything else.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      q_list (list): list of dictionaries of the possible NQR frequencies in Hz
                        The keys of the dictionary are the possible m->m+1 values
                        For example: "m=1->2" for non-quadrupole active nuclei
                        the corresponding element will be an empty dictionary.
    """

    default_name = "efg_nqr"
    default_params = {
        "force_recalc": False,
        "use_q_isotopes": False,
        "isotopes": {},
        "isotope_list": None,
        "tag": DEFAULT_EFG_TAG,
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, use_q_isotopes, isotopes, isotope_list, tag):


        # First thing, build the isotope dictionary
        elems = s.get_chemical_symbols()

        # Is isotope list valid?
        if isotope_list is not None and len(isotope_list) != len(elems):
            print("WARNING - invalid isotope_list, ignoring")
            isotope_list = None

        q_list = _get_isotope_data(elems, "Q", isotopes, isotope_list, use_q_isotopes)
        I_list = _get_isotope_data(elems, "I", isotopes, isotope_list, use_q_isotopes)
        eta_list = EFGAsymmetry.get(s, tag=tag, force_recalc=force_recalc)
        nqr = [defaultdict(None) for i in range(len(elems))]

        mask = I_list > 0.5
        for i in np.where(mask)[0]:
            A = EFG_TO_CHI * EFGVzz.get(s, tag=tag, force_recalc=force_recalc)[i] * q_list[i] / (4 * I_list[i] * (2 * I_list[i] - 1))
            ms = [m for m in _frange(-I_list[i], I_list[i] + 1, 1) if m >= 0.0][:-1]
            for m in ms:
                key = f'm={m}->{m+1}'
                fq = 3 * A * (2 * m + 1) * np.sqrt(1 + eta_list[i] ** 2 / 3)
                nqr[i][key] = fq


        return nqr

    @tensor_mean_property('NQR')
    def mean(self, s, axis=None, weights=None, **kwargs):
        """
        Calculate the mean of the EFGNQR property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.
          **kwargs: Additional parameters, including isotopes and use_q_isotopes.

        Returns:
          efg_nqr_mean (list): The mean of the EFGNQR property frequencies.
        """
        # Implementation handled by decorator


class EFGQuadrupolarProduct(AtomsProperty):

    """
    EFGQuadrupolarProduct

    Produces an array containing the quadrupolar product values
    in a system.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    .. math::

        \\P_Q = C_Q (1+frac{\\eta_Q^2}{3})^{1/2}

    Parameters:
      force_recalc (bool): if True, always diagonalise the tensors even if
                           already present.
      use_q_isotopes (bool): if True, always use the most common quadrupole
                             active isotope for each element, if there is
                             one.
      isotopes (dict): dictionary of specific isotopes to use, by element
                       symbol. If the isotope doesn't exist an error will
                       be raised. Overrides use_q_isotopes.
      isotope_list (list): list of isotopes, atom-by-atom. To be used if
                           different atoms of the same element are supposed
                           to be of different isotopes. Where a 'None' is
                           present will fall back on the previous
                           definitions. Where an isotope is present it
                           overrides everything else.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      Pq_list (np.ndarray): list of quadrupole products in Hz (units of Cq)
    """

    default_name = "efg_qprod"
    default_params = {
        "force_recalc": False,
        "use_q_isotopes": False,
        "isotopes": {},
        "isotope_list": None,
        "tag": DEFAULT_EFG_TAG,
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, use_q_isotopes, isotopes, isotope_list, tag):


        # First thing, build the isotope dictionary
        elems = s.get_chemical_symbols()

        # Is isotope list valid?
        if isotope_list is not None and len(isotope_list) != len(elems):
            print("WARNING - invalid isotope_list, ignoring")
            isotope_list = None

        q_list = _get_isotope_data(elems, "Q", isotopes, isotope_list, use_q_isotopes)

        return (
            EFG_TO_CHI
            * q_list
            * EFGVzz.get(s, tag=tag, force_recalc=force_recalc)
            * (1 + (EFGAsymmetry.get(s, tag=tag, force_recalc=force_recalc) ** 2) / 3) ** 0.5
        )

    @tensor_mean_property('Pq')
    def mean(self, s, axis=None, weights=None, **kwargs):
        """
        Calculate the mean of the EFGQuadrupolarProduct property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.
          **kwargs: Additional parameters, including isotopes and use_q_isotopes.

        Returns:
          efg_qprod_mean (np.ndarray): The mean of the EFGQuadrupolarProduct property.
        """
        # Implementation handled by decorator


class EFGEuler(AtomsProperty):

    """
    EFGEuler

    Produces an array of Euler angles in radians expressing the orientation of
    the EFG tensors with respect to the cartesian axes for each site in the Atoms object.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.


    Parameters:
        order (str):  Order to use for eigenvalues/eigenvectors. Can
                        be 'i' (ORDER_INCREASING), 'd'
                        (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                        'n' (ORDER_NQR). Default is 'n' for EFG tensors.
        convention (str): 'zyz' or 'zxz' accepted - the ordering of the Euler
                        angle rotation axes. Default is ZYZ 
        passive (bool):  active or passive rotations. Default is active (passive=False)
        tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                             Can be customized for non-standard tensor arrays.

    Returns:
        efg_eulers (np.array): array of Euler angles in radians
    """

    default_name = "efg_eulers"
    default_params = {"order": ElectricFieldGradient.ORDER_NQR,
                      "convention": "zyz",
                      "passive": False,
                      "tag": DEFAULT_EFG_TAG}

    @staticmethod
    @_has_efg_check
    def extract(s, order, convention, passive, tag):
        return np.array([t.euler_angles(convention, passive=passive) for t in EFGTensor.get(s, order=order, tag=tag)])

    def mean(self, s, axis=None, weights=None, **kwargs):
        """
        Calculate the mean of the EFGEuler property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.
          **kwargs: Additional parameters for euler angles calculation.

        Returns:
          efg_euler_mean (np.ndarray): The mean of the EFGEuler property.
        """
        # Get the mean EFGTensor
        meanTensors = EFGTensor().mean(s, axis=axis, weights=weights, **kwargs)
        # Get the Euler angles
        convention = kwargs.get('convention', 'zyz')
        passive = kwargs.get('passive', False)
        return np.array([t.euler_angles(convention=convention, passive=passive) for t in meanTensors])


class EFGQuaternion(AtomsProperty):

    """
    EFGQuaternion

    Produces a list of ase.Quaternion objects expressing the orientation of
    the EFG tensors with respect to the cartesian axes.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    The default convention for EFG tensors is the NQR one (`|Vxx| ≤ |Vyy| ≤ |Vzz|`).
    This is different from the default convention for MS tensors (Haeberlen).
    You can change this by specifying the 'order' parameter.

    Parameters:
      order (str):  Order to use for eigenvalues/eigenvectors. Can
                    be 'i' (ORDER_INCREASING), 'd'
                    (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                    'n' (ORDER_NQR). Default is 'n'.
      tag (str, optional): name of the array containing electric field gradient tensors. Defaults to 'efg'.
                           Can be customized for non-standard tensor arrays.

    Returns:
      efg_quat (list): list of quaternions
    """

    default_name = "efg_quats"
    default_params = {"order": ElectricFieldGradient.ORDER_NQR, "tag": DEFAULT_EFG_TAG}

    @staticmethod
    @_has_efg_check
    def extract(s, order, tag):
        return [t.quaternion for t in EFGTensor.get(s, order=order, tag=tag)]

    @tensor_mean_property('quaternion')
    def mean(self, s, axis=None, weights=None):
        """
        Calculate the mean of the EFGQuaternion property.

        Parameters:
          s (AtomsCollection): The collection of structures to calculate the mean for.
          axis (int or None): Axis along which to calculate the mean. Default is None.
          weights (array-like or None): Weights for each structure. Default is None.

        Returns:
          efg_quat_mean (np.ndarray): The mean of the EFGQuaternion property.
        """
        # Implementation handled by decorator



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes and functions for simulating approximated NMR spectroscopic results
from structures.
"""


import logging
import re
from collections import namedtuple
from collections.abc import Iterable
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from adjustText import adjust_text
from ase import Atoms
from matplotlib.axes import Axes
from scipy.special import fresnel

from soprano.calculate.nmr.utils import (
    Peak2D,
    calculate_distances,
    extract_indices,
    filter_pairs_by_distance,
    generate_contour_map,
    generate_peaks,
    get_atom_labels,
    get_pair_dipolar_couplings,
    get_pair_j_couplings,
    merge_peaks,
    nmr_2D_style,
    nmr_base_style,
    prepare_species_labels,
    process_pairs,
    sort_peaks,
    styled_plot,
    validate_elements,
)
from soprano.calculate.powder.triavg import TriAvg
from soprano.data.nmr import EFG_TO_CHI, _el_iso, _get_isotope_data, _get_isotope_list, _get_nmr_data
from soprano.nmr.utils import _dip_constant
from soprano.properties.nmr import MSIsotropy
from soprano.selection import AtomSelection
from soprano.utils import minimum_supcell, supcell_gridgen

DEFAULT_MARKER_SIZE = 50 # Default marker size for 2D plots (max is 100)
ANNOTATION_LINE_WIDTH = 0.15 # Line width for annotation lines
ANNOTATION_FONT_SCALE = 0.5 # Scale factor for annotation font size wrt to plot font size
DEFAULT_MAX_NUM_LEGEND_ELEMENTS = 6 # Default maximum number of elements to show in the legend in 2D plots

_nmr_data = _get_nmr_data()

# Conversion functions to Tesla
# (they take element and isotope as arguments)
_larm_units = {
    "MHz": lambda e, i: 2 * np.pi * 1.0e6 / _nmr_data[e][i]["gamma"],
    "T": lambda e, i: 1.0,
}

# Function used for second-order quadrupolar shift
# Arguments: cos(2*alpha), eta


def _st_A(c2a, eta):
    return -27.0 / 8.0 - 9.0 / 4.0 * eta * c2a - 3.0 / 8.0 * eta ** 2 * c2a ** 2


def _st_B(c2a, eta):
    return 15.0 / 4.0 - 0.5 * eta ** 2 + 2 * eta * c2a + 0.75 * eta ** 2 * c2a ** 2


def _st_C(c2a, eta):
    return (
        -23.0 / 40.0
        + 14.0 / 15.0 * eta ** 2
        + 0.25 * eta * c2a
        - 3.0 / 8.0 * eta ** 2 * c2a ** 2
    )


def _mas_A(c2a, eta):
    return 21.0 / 16.0 - 7.0 / 8.0 * eta * c2a + 7.0 / 48.0 * eta ** 2 * c2a ** 2


def _mas_B(c2a, eta):
    return (
        -9.0 / 8.0
        + 1.0 / 12.0 * eta ** 2
        + eta * c2a
        - 7.0 / 24.0 * eta ** 2 * c2a ** 2
    )


def _mas_C(c2a, eta):
    return (
        9.0 / 80.0
        - 1.0 / 15.0 * eta ** 2
        - 0.125 * eta * c2a
        + 7.0 / 48.0 * eta ** 2 * c2a ** 2
    )


def _gfunc(ca, cb, eta, A, B, C):
    c2a = 2 * ca ** 2 - 1
    return A(c2a, eta) * cb ** 4 + B(c2a, eta) * cb ** 2 + C(c2a, eta)


# Flags for what to include in spectra
NMRFlags = namedtuple(
    "NMRFlags",
    """CS_ISO
                      CS_ORIENT
                      CS
                      Q_1_ORIENT
                      Q_2_SHIFT
                      Q_2_ORIENT_STATIC
                      Q_2_ORIENT_MAS
                      Q_2_STATIC
                      Q_2_MAS
                      Q_STATIC
                      Q_MAS
                      STATIC
                      MAS""",
)
NMRFlags = NMRFlags(
    CS_ISO=1,
    CS_ORIENT=2,
    CS=1 + 2,
    Q_1_ORIENT=4,
    Q_2_SHIFT=8,
    Q_2_ORIENT_STATIC=16,
    Q_2_ORIENT_MAS=32,
    Q_2_STATIC=8 + 16,
    Q_2_MAS=8 + 32,
    Q_STATIC=4 + 8 + 16,
    Q_MAS=8 + 32,
    STATIC=1 + 2 + 4 + 8 + 16,
    MAS=1 + 8 + 32,
)


MARKER_INFO = {
    'distance': {
        'label': 'Distance',
        'unit': 'Å',
        'fmt': '{x:.1f}'
    },
    'inversedistance': {
        'label': '1/Distance',
        'unit': r'Å$^{{-1}}$',
        'fmt': '{x:.3f}'
    },
    'dipolar': {
        'label': 'Dipolar Coupling',
        'unit': 'kHz',
        'fmt': '{x:.1f}'
    },
    'jcoupling': {
        'label': 'J Coupling',
        'unit': 'Hz',
        'fmt': '{x:.1f}'
    },
    'fixed': {
        'label': 'Fixed',
        'unit': '',
        'fmt': '{x:.1f}'
    },
    'custom': {
        'label': 'Correlation strength',
        'unit': '',
        'fmt': '{x:.1f}'
    }
    }

class NMRCalculator:

    """NMRCalculator

    An object providing an interface to produce basic simulated NMR spectra
    from .magres files. It should be kept in mind that this is *not* a proper
    spin simulation tool, but merely provides a 'guide for the eye' kind of
    spectrum to compare to experimental results. What it can simulate:

    - chemical shift of NMR peaks
    - quadrupolar shifts of NMR peaks up to second order corrections
    - effects of crystal orientation (single crystal)
    - powder average (policrystalline/powder)
    - ultrafast MAS limit spectra

    What it can NOT simulate:

    - finite speed MAS spectra
    - J couplings
    - dipolar interactions
    - complex NMR experiments

    A list of the currently available NMRFlags to be used in conjunction with
    methods that require a list of effects of interest: ::

        NMRFlags.CS_ISO     => chemical shielding, isotropic effect
                .CS_ORIENT  => chemical shielding, orientation dependent
                               effects
                .CS         => chemical shielding, everything
                .Q_1_ORIENT => quadrupolar, 1st order, orientation dependent
                               effects
                .Q_2_SHIFT  => quadrupolar, 2nd order, isotropic shift
                .Q_2_ORIENT_STATIC => quadrupolar, 2nd order, orientation
                                      dependent effects; static limit
                .Q_2_ORIENT_MAS => quadrupolar, 2nd order, orientation
                                      dependent effects; ultrafast MAS limit
                .Q_2_STATIC => quadrupolar, 2nd order, all static effects
                .Q_2_MAS    => quadrupolar, 2nd order, all ultrafast MAS
                               effects
                .Q_STATIC   => quadrupolar, all static effects
                .Q_MAS      => quadrupolar, all ultrafast MAS effects
                .STATIC     => all static effects
                .MAS        => all ultrafast MAS effects


    | Args:
    |   sample (ase.Atoms): an Atoms object describing the system to simulate
    |                       on. Should be loaded with ASE from a .magres file
    |                       if data on shieldings and EFGs is necessary. It
    |                       can also have an optional 'isotopes' array. If it
    |                       does, it will be used in the set_isotopes method
    |                       and interpreted as described in its documentation.
    |   larmor_frequency (float): larmor frequency of the virtual spectrometer
    |                             (referenced to Hydrogen). Default is 400.
    |   larmor_units (str): units in which the larmor frequency is expressed.
    |                       Default are MHz.

    """

    def __init__(self, sample, larmor_frequency=400, larmor_units="MHz"):

        if not isinstance(sample, Atoms):
            raise TypeError("sample must be an ase.Atoms object")

        self._sample = sample

        # Define isotope array
        self._elems = np.array(self._sample.get_chemical_symbols())
        if self._sample.has("isotopes"):
            isos = self._sample.get_array("isotopes")
        else:
            isos = [None] * len(self._sample)
        self.set_isotopes(isos)

        self.set_larmor_frequency(larmor_frequency, larmor_units)

        self._references = {}

    def set_larmor_frequency(
        self, larmor_frequency=400, larmor_units="MHz", element="1H"
    ):
        """
        Set the Larmor frequency of the virtual spectrometer with the desired
        units and reference element.

        | Args:
        |   larmor_frequency (float): larmor frequency of the virtual
        |                             spectrometer. Default is 400.
        |   larmor_units (str): units in which the larmor frequency is
        |                       expressed. Can be MHz or T. Default are MHz.
        |   element (str): element and isotope to reference the frequency to.
        |                  Should be in the form <isotope><element>. Isotope
        |                  is optional, if absent the most abundant NMR active
        |                  one will be used. Default is 1H.

        """

        if larmor_units not in _larm_units:
            raise ValueError("Invalid units for Larmor frequency")

        # Split isotope and element
        el, iso = _el_iso(element)

        self._B = larmor_frequency * _larm_units[larmor_units](el, iso)

    def get_larmor_frequency(self, element):
        """
        Get the Larmor frequency of the virtual spectrometer for the desired
        element in MHz.

        | Args:
        |   element (str): element and isotope whose frequency we require.
        |                  Should be in the form <isotope><element>. Isotope
        |                  is optional, if absent the most abundant NMR active
        |                  one will be used. Default is 1H.

        | Returns:
        |   larmor (float): Larmor frequency in MHz

        """

        el, iso = _el_iso(element)
        return self._B / _larm_units["MHz"](el, iso)

    def set_reference(self, ref, element):
        """
        Set the chemical shift reference (in ppm) for a given element. If not
        provided it will be assumed to be zero.

        | Args:
        |   ref (float): reference shielding value in ppm. Chemical shift will
        |                be calculated as this minus the atom's ms.
        |   element (str): element and isotope whose reference is set.
        |                  Should be in the form <isotope><element>. Isotope
        |                  is optional, if absent the most abundant NMR active
        |                  one will be used.

        """

        el, iso = _el_iso(element)

        if el not in self._references:
            self._references[el] = {}
        self._references[el][iso] = float(ref)

    def set_isotopes(self, isotopes):
        """
        Set the isotopes for each atom in sample.

        | Args:
        |   isotopes (list): list of isotopes for each atom in sample.
        |                    Isotopes can be given as an array of integers or
        |                    of symbols in the form <isotope><element>.
        |                    Their order must match the one of the atoms in
        |                    the original sample ase.Atoms object.
        |                    If an element of the list is None, the most
        |                    common NMR-active isotope is used. If an element
        |                    is the string 'Q', the most common quadrupolar
        |                    active isotope for that nucleus (if known) will
        |                    be used.

        """

        # First: correct length?
        if len(isotopes) != len(self._sample):
            raise ValueError(
                "isotopes array should be as long as the atoms" " in sample"
            )

        # Clean up the list, make sure it's all right
        iso_clean = []
        for i, iso in enumerate(isotopes):
            # Is it an integer?
            iso_name = ""
            if re.match("[0-9]+", str(iso)) is not None:  # numpy-proof test
                iso_name = str(iso)
                # Does it exist?
                if iso_name not in _nmr_data[self._elems[i]]:
                    raise ValueError(
                        "Invalid isotope "
                        f"{iso_name} for element {self._elems[i]}"
                    )
            elif iso is None:
                iso_name = str(_nmr_data[self._elems[i]]["iso"])
            elif iso == "Q":
                iso_name = str(_nmr_data[self._elems[i]]["Q_iso"])
            else:
                el, iso_name = _el_iso(iso)
                # Additional test
                if el != self._elems[i]:
                    raise ValueError(
                        "Invalid element in isotope array - "
                        f"{el} in place of {self._elems[i]}"
                    )

            iso_clean.append(iso_name)

        self._isos = np.array(iso_clean)

    def set_element_isotope(self, element, isotope):
        """
        Set the isotope for all occurrences of a given element.

        | Args:
        |   element (str): chemical symbol of the element for which to set the
        |                  isotope.
        |   isotope (int or str): isotope to set for the given element. The
        |                         same conventions as described for the array
        |                         passed to set_isotopes apply.

        """

        new_isos = [int(x) for x in self._isos]
        new_isos = np.where(self._elems == element, isotope, new_isos)

        self.set_isotopes(new_isos)

    def set_single_crystal(self, theta, phi):
        """
        Set the orientation of the sample as a single crystallite.

        | Args:
        |   theta (float): zenithal angle for the crystallite
        |   phi (float): azimuthal angle for the crystallite

        """

        p = np.array(
            [[np.sin(theta) * np.cos(phi), np.sin(theta) * np.sin(phi), np.cos(theta)]]
        )
        w = np.array([1.0])
        t = np.array([])

        self._orients = [p, w, t]

    def set_powder(self, N=8, mode="hemisphere"):
        """
        Set the orientation of the sample as a powder average.

        | Args:
        |   N (int): the number of subdivisions used to generate orientations
        |            within the POWDER algorithm. Higher values make for
        |            better but more expensive averages.
        |   mode (str): which part of the solid angle to cover with the
        |               orientations. Can be 'octant', 'hemisphere' or
        |               'sphere'. The latter should not be necessary for any
        |               NMR interaction. Default is 'hemisphere'.

        """

        self._pwdscheme = TriAvg(mode)
        self._orients = self._pwdscheme.get_orient_points(N)

    def spectrum_1d(
        self,
        element,
        min_freq=-50,
        max_freq=50,
        bins=100,
        freq_broad=None,
        freq_units="ppm",
        effects=NMRFlags.CS_ISO,
        use_central=False,
        use_reference: Optional[bool]=None,
    ):
        """
        Return a simulated spectrum for the given sample and element.

        | Args:
        |   element (str): element and isotope to get the spectrum of.
        |                  Should be in the form <isotope><element>. Isotope
        |                  is optional, if absent the most abundant NMR active
        |                  one will be used.
        |   min_freq (float): lower bound of the frequency range
        |                    (default is -50)
        |   min_freq (float): upper bound of the frequency range
        |                    (default is 50)
        |   bins (int): number of bins in which to separate the frequency range
        |              (default is 500)
        |   freq_broad (float): Gaussian broadening width to apply to the
        |                       final spectrum (default is None)
        |   freq_units (str): units used for frequency, can be ppm or MHz
        |                     (default is ppm).
        |   effects (NMRFlags): a flag, or bitwise-joined set of flags, from
        |                       this module's NMRFlags tuple, describing which
        |                       effects should be included and accounted for
        |                       in the calculation. For a list of available
        |                       flags check the docstring for NMRCalculator
        |                       (default is NMRFlags.CS_ISO).
        |   use_central (bool): if True, for half-integer spin nuclei, only
        |                       show the central transition. Ignored for
        |                       integer spin nuclei (default is False).
        |   use_reference (bool): if True, return frequencies as referenced to
        |                         the appropriate nucleus, in chemical shift
        |                         form. If no reference has been provided for
        |                         this nucleus, a value of 0 ppm is used and
        |                         the frequencies are simply flipped in sign
        |                         (default is None, in which case use_reference is True if
        |                          references have been set).

        | Returns:
        |   spec (np.ndarray): array of length 'bins' containing the spectral
        |                      intensities
        |   freq (np.ndarray): array of length 'bins' containing the frequency
        |                      axis

        """

        # First, define the frequency range
        el, iso = _el_iso(element)
        larm = self._B * _nmr_data[el][iso]["gamma"] / (2.0 * np.pi * 1e6)
        I = _nmr_data[el][iso]["I"]
        # Units? We want this to be in ppm
        u = {
            "ppm": 1,
            "MHz": 1e6 / larm,
        }
        try:
            freq_axis = np.linspace(min_freq, max_freq, bins) * u[freq_units]
        except KeyError:
            raise ValueError("Invalid freq_units passed to spectrum_1d")

        # If it's not a quadrupolar nucleus, no reason to keep those effects
        # around...
        if abs(I) < 1:
            effects &= ~NMRFlags.Q_STATIC
            effects &= ~NMRFlags.Q_MAS

        # Ok, so get the relevant atoms and their properties
        a_inds = np.where((self._elems == el) & (self._isos == iso))[0]

        # Are there even any such atoms?
        if len(a_inds) == 0:
            raise RuntimeError(
                "No atoms of the desired isotopes found in the" " system"
            )

        # Sanity check
        if effects & NMRFlags.Q_2_ORIENT_STATIC and effects & NMRFlags.Q_2_ORIENT_MAS:
            # Makes no sense...
            raise ValueError(
                "The flags Q_2_ORIENT_STATIC and Q_2_ORIENT_MAS"
                " can not be set at the same time"
            )

        if effects & NMRFlags.CS:
            try:
                ms_tens = self._sample.get_array("ms")[a_inds]
                ms_tens = (ms_tens + np.swapaxes(ms_tens, 1, 2)) / 2.0
            except KeyError:
                raise RuntimeError(
                    "Impossible to compute chemical shift - "
                    "sample has no shielding data"
                )
            ms_evals, ms_evecs = zip(*[np.linalg.eigh(t) for t in ms_tens])

        if effects & (NMRFlags.Q_STATIC | NMRFlags.Q_MAS):
            try:
                efg_tens = self._sample.get_array("efg")[a_inds]
            except KeyError:
                raise RuntimeError(
                    "Impossible to compute quadrupolar effects"
                    " - sample has no EFG data"
                )
            efg_evals, efg_evecs = zip(*[np.linalg.eigh(t) for t in efg_tens])
            efg_i = (
                np.arange(len(efg_evals))[:, None],
                np.argsort(np.abs(efg_evals), axis=1),
            )
            efg_evals = np.array(efg_evals)[efg_i]
            efg_evecs = np.array(efg_evecs)[efg_i[0], :, efg_i[1]]
            Vzz = efg_evals[:, -1]
            eta_q = (efg_evals[:, 0] - efg_evals[:, 1]) / Vzz
            Q = _nmr_data[el][iso]["Q"]
            chi = Vzz * Q * EFG_TO_CHI

        # Reference (zero if not given)
        try:
            ref = self._references[el][iso]
        except KeyError:
            ref = 0.0

        # Default to using reference if it's not zero
        if use_reference is None:
            use_reference = ref != 0.0

        # Let's start with peak positions - quantities non dependent on
        # orientation

        # Shape: atoms*1Q transitions
        peaks = np.zeros((len(a_inds), int(2 * I)))

        # Magnetic quantum number values
        if I % 1 == 0.5 and use_central:
            m = np.array([-0.5, 0.5])[None, :]
        else:
            m = np.arange(-I, I + 1).astype(float)[None, :]

        if effects & NMRFlags.CS_ISO:
            peaks += np.average(ms_evals, axis=1)[:, None]

        # Quadrupole second order
        if effects & NMRFlags.Q_2_SHIFT:
            nu_l = larm * 1e6
            # NOTE: the last factor of two in this formula was inserted
            # despite not being present in M. J. Duer (5.9) as apparently
            # it's a mistake in the book. Other sources (like the quadrupolar
            # NMR online book by D. Freude and J. Haase, Dec. 2016) report
            # this formulation, with the factor of two, instead.
            q_shifts = np.diff(
                (chi[:, None] / (4 * I * (2 * I - 1))) ** 2
                * m
                / nu_l
                * (-0.2 * (I * (I + 1) - 3 * m ** 2) * (3 + eta_q[:, None] ** 2))
                * 2
            )
            q_shifts /= larm

            peaks += q_shifts

        # Any orientational effects at all?
        has_orient = effects & (
            NMRFlags.CS_ORIENT
            | NMRFlags.Q_1_ORIENT
            | NMRFlags.Q_2_ORIENT_STATIC
            | NMRFlags.Q_2_ORIENT_MAS
        )
        # Are we using a POWDER average?
        use_pwd = len(self._orients[2]) > 0

        if has_orient:
            # Further expand the peaks!
            peaks = np.repeat(peaks[:, :, None], len(self._orients[0]), axis=-1)

        # Now compute the orientational quantities
        if effects & NMRFlags.CS_ORIENT:

            # Compute the traceless ms tensors
            ms_traceless = ms_tens - [
                np.identity(3) * np.average(ev) for ev in ms_evals
            ]
            # Now get the shift contributions for each orientation
            dirs = self._orients[0]

            peaks += np.sum(
                dirs.T[None, :, :] * np.tensordot(ms_traceless, dirs, axes=((2), (1))),
                axis=1,
            )[:, None, :]

        if effects & NMRFlags.Q_1_ORIENT:

            # First order quadrupolar anisotropic effects
            # We consider the field aligned along Z
            cosb2 = self._orients[0][:, 2] ** 2
            sinb2 = 1.0 - cosb2
            cosa2 = (self._orients[0][:, 0] ** 2) / np.where(sinb2 > 0, sinb2, np.inf)

            dir_fac = 0.5 * (
                (3 * cosb2[None, :] - 1)
                + eta_q[:, None] * sinb2[None, :] * (2 * cosa2[None, :] - 1.0)
            )
            m_fac = m[:, :-1] + 0.5
            nu_q = chi * 1.5 / (I * (2 * I - 1.0))

            qfreqs = nu_q[:, None, None] * m_fac[:, :, None] * dir_fac[:, None, :]

            peaks += qfreqs / larm  # Already ppm being Hz/MHz

        if effects & (NMRFlags.Q_2_ORIENT_STATIC | NMRFlags.Q_2_ORIENT_MAS):
            # Which one?
            if effects & NMRFlags.Q_2_ORIENT_STATIC:
                ABC = [_st_A, _st_B, _st_C]
            else:
                ABC = [_mas_A, _mas_B, _mas_C]

            cosa = self._orients[0][:, 0]
            cosb = self._orients[0][:, 1]

            dir_fac = _gfunc(cosa[None, :], cosb[None, :], eta_q[:, None], *ABC)

            m_fac = (
                I * (I + 1.0) - 17.0 / 3.0 * m[:, :-1] * (m[:, :-1] + 1) - 13.0 / 6.0
            )
            nu_q = chi * 1.5 / (I * (2 * I - 1.0))

            qfreqs = -(
                (nu_q ** 2 / (6.0 * larm * 1e6))[:, None, None]
                * m_fac[:, :, None]
                * dir_fac[:, None, :]
            )

            peaks += qfreqs / larm

        # Finally, the overall spectrum
        spec = np.zeros(freq_axis.shape)

        for p_nuc in peaks:
            for p_trans in p_nuc:
                if has_orient and use_pwd:
                    spec += self._pwdscheme.average(
                        freq_axis, p_trans, self._orients[1], self._orients[2]
                    )

        if freq_broad is None and (not has_orient or not use_pwd):
            print(
                "WARNING: no artificial broadening detected in a calculation"
                " without line-broadening contributions. The spectrum could "
                "appear distorted or empty"
            )

        if freq_broad is not None:
            if has_orient and use_pwd:
                fc = (max_freq + min_freq) / 2.0
                bk = np.exp(-(((freq_axis - fc) / freq_broad) ** 2.0))
                bk /= np.sum(bk)
                spec = np.convolve(spec, bk, mode="same")
            else:
                bpeaks = np.exp(
                    -(((freq_axis - peaks[:, :, None]) / freq_broad) ** 2)
                )  # Broadened peaks
                # Normalise them BY PEAK MAXIMUM
                norm_max = np.amax(bpeaks, axis=-1, keepdims=True)
                norm_max = np.where(np.isclose(norm_max, 0), np.inf, norm_max)
                bpeaks /= norm_max
                spec = np.sum(bpeaks, axis=(0, 1) if not has_orient else (0, 1, 2))

        # Normalize the spectrum to the number of nuclei
        normsum = np.sum(spec)
        if np.isclose(normsum, 0):
            print(
                "WARNING: no peaks found in the given frequency range. "
                "The spectrum will be empty"
            )
        else:
            spec *= len(a_inds) * len(spec) / normsum

        if use_reference:
            freq_axis = ref - freq_axis

        freqs = freq_axis / u[freq_units]
        return spec, freqs

    def dq_buildup(
        self,
        sel_i,
        sel_j=None,
        t_max=1e-3,
        t_steps=1000,
        R_cut=3,
        kdq=0.155,
        A=1,
        tau=np.inf,
    ):
        """
        Return a dictionary of double quantum buildup curves for given pairs
        of atoms, built according to the theory given in:

        G. Pileio et al., "Analytical theory of gamma-encoded double-quantum
        recoupling sequences in solid-state nuclear magnetic resonance"
        Journal of Magnetic Resonance 186 (2007) 65-74

        | Args:
        |   sel_i (AtomSelection or [int]): Selection or list of indices of
        |                                   atoms for which to compute the
        |                                   curves. By default is None
        |                                   (= all of them).
        |   sel_i (AtomSelection or [int]): Selection or list of indices of
        |                                   atoms for which to compute the
        |                                   curves with sel_i. By default is
        |                                   None (= same as sel_i).
        |   t_max (float): maximum DQ buildup time, in seconds. Default
        |                  is 1e-3.
        |   t_steps (int): number of DQ buildup time steps. Default is 1000.
        |   R_cut (float): cutoff radius for which periodic copies to consider
        |                  in each pair, in Angstrom. Default is 3.
        |   kdq (float): same as the k constant in eq. 35 of the reference. A
        |                parameter depending on the specific sequence used.
        |                Default is 0.155.
        |   A (float): overall scaling factor for the curve. Default is 1.
        |   tau (float): exponential decay factor for the curve. Default
        |                is np.inf.

        | Returns:
        |   curves (dict): a dictionary of all buildup curves indexed by pair,
        |                  plus the time axis in seconds as member 't'.
        """

        tdq = np.linspace(0, t_max, t_steps)

        # Selections
        if sel_i is None:
            sel_i = AtomSelection.all(self._sample)
        elif not isinstance(sel_i, AtomSelection):
            sel_i = AtomSelection(self._sample, sel_i)

        if sel_j is None:
            sel_j = sel_i
        elif not isinstance(sel_j, AtomSelection):
            sel_j = AtomSelection(self._sample, sel_j)

        # Find gammas
        elems = self._sample.get_chemical_symbols()
        gammas = _get_isotope_data(elems, "gamma", {}, self._isos)

        # Need to sort them and remove any duplicates, also take i < j as
        # convention
        pairs = [
            tuple(sorted((i, j)))
            for i in sorted(sel_i.indices)
            for j in sorted(sel_j.indices)
        ]

        scell_shape = minimum_supcell(R_cut, latt_cart=self._sample.get_cell())
        nfg, ng = supcell_gridgen(self._sample.get_cell(), scell_shape)

        pos = self._sample.get_positions()

        curves = {"t": tdq}

        for ij in pairs:

            r = pos[ij[1]] - pos[ij[0]]

            all_r = r[None, :] + ng

            all_R = np.linalg.norm(all_r, axis=1)

            # Apply cutoff
            all_R = all_R[np.where((all_R <= R_cut) * (all_R > 0))]

            bij = _dip_constant(all_R * 1e-10, gammas[ij[0]], gammas[ij[1]])

            th = 1.5 * kdq * abs(bij[:, None]) * tdq[None, :] * 2 * np.pi
            x = (2 * th / np.pi) ** 0.5

            Fs, Fc = fresnel(x * 2 ** 0.5)

            x[:, 0] = np.inf

            bdup = 0.5 - (1.0 / (x * 8 ** 0.5)) * (
                Fc * np.cos(2 * th) + Fs * np.sin(2 * th)
            )
            bdup[:, 0] = 0

            curves[ij] = A * np.sum(bdup, axis=0) * np.exp(-tdq / tau)

        return curves

    @property
    def B(self):
        """Static magnetic field, in Tesla"""
        return self._B


class NMRData2D:
    '''
    Class to hold and extract 2D NMR data.
    '''
    def __init__(self,
                atoms: Optional[Atoms] = None,
                xelement: Optional[str] = None,
                yelement: Optional[str] = None,
                references: Optional[dict[str, float]] = None,
                gradients: Optional[dict[str, float]] = None,
                peaks: Optional[List[Peak2D]] = None,
                pairs: Optional[List[Tuple[int, int]]] = None,
                correlation_strengths: Optional[List[float]] = None,
                correlation_strength_metric: Optional[str] = None, # 'fixed','distance', 'dipolar', 'jcoupling', 'inversedistance', 'custom'
                rcut: Optional[float] = None,
                isotopes: Optional[dict[str, int]] = None,
                is_shift: Optional[bool] = None,
                include_quadrupolar: bool = False,
                yaxis_order: str = '1Q',
                x_axis_label: Optional[str] = None,
                y_axis_label: Optional[str] = None,
                ):

        self.atoms = atoms
        self.xelement = xelement
        # if yelement is not provided, set it to xelement
        self.yelement = yelement if yelement is not None else xelement
        self.references = references
        self.gradients = gradients
        self.peaks = peaks
        self.pairs = pairs

        # if neither atoms nor peaks are provided, raise an error
        if self.atoms is None and self.peaks is None:
            raise ValueError("Either atoms or peaks must be given.")


        # Either provide correlation strengths or calculate them based on the metric
        # If both are provided, the provided values will be used
        if correlation_strengths:
            if correlation_strength_metric:
                self.logger.warning("Both correlation_strengths and correlation_strength_metric are provided. Using correlation_strengths.")
            correlation_strength_metric = 'custom'


        # If neither are provided, set the metric to 'fixed'
        if correlation_strengths is None and correlation_strength_metric is None:
            correlation_strength_metric = 'fixed'

        self.correlation_strengths = correlation_strengths
        self.correlation_strength_metric = correlation_strength_metric

        self.rcut = rcut
        self.isotopes = isotopes if isotopes is not None else {}
        # is_shift is a boolean.  If undefined, it will be set to True if references are provided, False otherwise
        #  If defined, it will be used as is
        self.is_shift = is_shift if is_shift is not None else (self.references is not None)
        self.include_quadrupolar = include_quadrupolar
        self.yaxis_order = yaxis_order
        self.logger = logging.getLogger(__name__)

        self.correlation_unit = MARKER_INFO[self.correlation_strength_metric]['unit']
        self.correlation_label = MARKER_INFO[self.correlation_strength_metric]['label']
        self.correlation_fmt = MARKER_INFO[self.correlation_strength_metric]['fmt']

        self.x_axis_label = x_axis_label
        self.y_axis_label = y_axis_label

        # run the main method to extract the data
        self.get_peaks()

    def get_peaks(self, merge_identical=True, should_sort_peaks=False):
        '''
        Get the correlation peaks.

        If self.peaks already exists, then we return them as is.
        If they don't exist, we make sure the required data is available 
        and then generate the peaks and merge if desired.
        '''

        if self.peaks is not None:
            self.logger.debug("Custom peaks provided. ")
            return self.peaks

        if self.atoms is None:
            raise ValueError("Either atoms or peaks must be given.")

        # make sure all the data is there
        self.extract_data()
        labels = get_atom_labels(self.atoms, self.logger)

        self.peaks = generate_peaks(self.data, self.pairs, labels, self.correlation_strengths, self.yaxis_order, self.xelement, self.yelement)

        if merge_identical:
            self.peaks = merge_peaks(self.peaks)

        if should_sort_peaks:
            self.peaks = sort_peaks(self.peaks)

        return self.peaks

    def extract_data(self):
        validate_elements(self.atoms, self.xelement, self.yelement)
        self.idx_x, self.idx_y = extract_indices(self.atoms, self.xelement, self.yelement)
        isotopes = _get_isotope_list(self.atoms.get_chemical_symbols(), isotopes=self.isotopes, use_q_isotopes=False)
        self.xisotope = isotopes[self.idx_x[0]]
        self.yisotope = isotopes[self.idx_y[0]]
        self.xspecies = prepare_species_labels(self.xisotope, self.xelement)
        self.yspecies = prepare_species_labels(self.yisotope, self.yelement)

        self.get_axis_labels()
        self.get_pairs()
        if self.correlation_strength_metric is None:
            raise ValueError("correlation_strength_metric is not defined. Please provide a value for this parameter.")


        self.correlation_strengths = self.get_correlation_strengths()

        self.data = MSIsotropy.get(self.atoms, ref=self.references, grad=self.gradients)
        self.logger.debug(f'Indices of xelement in the atoms object: {self.idx_x}')
        self.logger.debug(f'Indices of yelement in the atoms object: {self.idx_y}')
        self.logger.debug(f'X species: {self.xspecies}')
        self.logger.debug(f'Y species: {self.yspecies}')
        self.logger.debug(f'X values: {self.data[self.idx_x]}')
        self.logger.debug(f'Y values: {self.data[self.idx_y]}')


    def get_correlation_strengths(self):
        if self.pairs is None:
            raise ValueError("No pairs found after filtering. Please check the input file and/or the user-specified filters.")

        if self.correlation_strength_metric == 'custom':
            self.logger.info("Using custom correlation strengths.")

            # make sure correlation_strengths is a list or an array
            if not isinstance(self.correlation_strengths, (list, np.ndarray)):
                raise TypeError("correlation_strengths must be a list or an array.")

            # if user provides a list of these, use it!
            # just check that it's the right length
            if len(self.correlation_strengths) != len(self.pairs_el_idx):
                raise ValueError(f"Length of correlation_strengths ({len(self.correlation_strengths)}) does not match the number of pairs ({len(self.pairs_el_idx)}).")
            correlation_strengths = self.correlation_strengths

        elif self.correlation_strength_metric == 'fixed':
            self.logger.info("Using fixed correlation strength.")
            # get all unique pairs of x and y indices
            # set the correlation strength to be the same for all pairs
            correlation_strengths = np.ones(len(self.pairs_el_idx))

        elif self.correlation_strength_metric == 'dipolar':
            self.logger.info("Using dipolar coupling as correlation strength.")
            if self.isotopes:
                self.logger.debug(f"Using custom isotopes: {self.isotopes}")
            correlation_strengths = get_pair_dipolar_couplings(self.atoms, self.pairs, self.isotopes)
        elif self.correlation_strength_metric == 'distance' or self.correlation_strength_metric == 'inversedistance':
            log_message = "Using minimum image convention {isinverse}distance as correlation strength."
            isinverse = ''

            # now we can use ASE get_distance to get the distances for each pair
            correlation_strengths = self.pair_distances
            if self.correlation_strength_metric == 'inversedistance':
                correlation_strengths = 1 / correlation_strengths
                isinverse = 'inverse '
            self.logger.info(log_message.format(isinverse=isinverse))


        elif self.correlation_strength_metric == 'jcoupling':
            self.logger.info("Using J-coupling as correlation strength.")
            correlation_strengths = get_pair_j_couplings(self.atoms, self.pairs, self.isotopes)
        elif self.correlation_strength_metric == 'custom':
            self.logger.info("Using custom correlation strengths.")
            correlation_strengths = self.correlation_strengths
        else:
            raise ValueError(f"Unknown correlation_strength_metric option: {self.correlation_strength_metric}")

        # Make sure correlation_strengths is an array
        correlation_strengths = np.array(correlation_strengths)

        self.logger.debug(f"correlation_strengths : {correlation_strengths}")
        # Log pair with smallest and largest correlation strength
        min_idx = np.argmin(np.abs(correlation_strengths))
        max_idx = np.argmax(np.abs(correlation_strengths))
        smallest_pair = self.pairs[min_idx]
        largest_pair = self.pairs[max_idx]

        # Labels for the smallest and largest pairs
        labels = get_atom_labels(self.atoms, self.logger)
        smallest_pair_labels = [labels[smallest_pair[0]], labels[smallest_pair[1]]]
        largest_pair_labels = [labels[largest_pair[0]], labels[largest_pair[1]]]

        self.logger.info(
            f"Pair with smallest (abs) {self.correlation_label}: "
            f"{smallest_pair_labels} ({correlation_strengths[min_idx]:.2f})"
        )
        self.logger.info(
            f"Pair with largest (abs) {self.correlation_label}: "
            f"{largest_pair_labels} ({correlation_strengths[max_idx]:.2f})"
        )

        return correlation_strengths

    def get_axis_labels(self):
        if self.is_shift:
            axis_label = r"$\delta$"
        else:
            axis_label = r"$\sigma$"

        if self.x_axis_label is None:
            self.x_axis_label = f'{self.xspecies} ' + axis_label + ' /ppm'
        if self.y_axis_label is None:
            if self.yaxis_order == '2Q':
                self.y_axis_label = f'{self.yspecies} ' + axis_label + r'$_{\mathrm{%s}}$' % self.yaxis_order + ' /ppm'
            else:
                self.y_axis_label = f'{self.yspecies} ' + axis_label + ' /ppm'


    def get_pairs(self):
        '''
        Get the pairs of x and y indices to plot

        self.idx_x and self.idx_y are the indices of the x and y elements in the atoms object

        This method will set the following attributes:
        self.pairs_el_idx: a list of tuples of the form (xindex, yindex)
        self.pairs: a list of tuples of the form (xindex, yindex)
        '''
        # Process pairs
        self.pairs, self.pairs_el_idx, self.idx_x, self.idx_y = process_pairs(self.idx_x, self.idx_y, self.pairs)

        # Check for invalid pairs if correlation_strength_metric is not fixed
        if self.correlation_strength_metric != 'fixed':
            for pair in self.pairs:
                if len(set(pair)) != 2:
                    raise ValueError("""
                    Two indices in a pair are the same but
                    the correlation_strength_metric is based on distance between sites.
                    It's unclear """)

        if len(self.pairs) == 0:
            raise ValueError("No pairs found after filtering. Please check the input file and/or the user-specified filters.")

        # Calculate distances
        # if pair_distances is not already calculated, calculate it
        if not hasattr(self, 'pair_distances'):
            self.pair_distances = calculate_distances(self.pairs, self.atoms)


        # Filter pairs by distance if rcut is specified
        if self.rcut:
            self.logger.info(f"Filtering out pairs that are further than {self.rcut} Å apart.")
            self.logger.info(f"Number of pairs before filtering: {len(self.pairs_el_idx)}")

            self.pairs, self.pairs_el_idx, self.pair_distances, self.idx_x, self.idx_y = filter_pairs_by_distance(
                self.pairs, self.pairs_el_idx, self.pair_distances, self.rcut)

            self.logger.info(f"Number of pairs remaining: {len(self.pairs_el_idx)}")
            self.logger.debug(f"Pairs remaining: {self.pairs}")
            self.logger.debug(f"Pairs el indices remaining: {self.pairs_el_idx}")




@dataclass
class PlotSettings:
    xlim: Optional[Tuple[float, float]] = None
    ylim: Optional[Tuple[float, float]] = None
    x_axis_label: Optional[str] = None
    y_axis_label: Optional[str] = None
    plot_filename: Optional[str] = None
    show_markers: bool = True
    marker: str = '+'
    marker_linewidth: float = 0.5
    max_marker_size: int = 10
    show_labels: bool = True
    auto_adjust_labels: bool = True
    label_fontsize: Optional[int] = None
    show_lines: bool = True
    show_diagonal: bool = True
    show_connectors: bool = True
    marker_color: Optional[str] = None
    show_legend: bool = False
    num_legend_elements: Optional[int] = None
    show_heatmap: bool = False
    heatmap_levels: Union[int, Iterable[float]] = 20
    show_contour: bool = False
    x_broadening: Optional[float] = None
    y_broadening: Optional[float] = None
    broadening_type: str = 'gaussian' # 'gaussian', 'lorentzian'
    heatmap_grid_size: int = 150
    colormap: str = 'bone_r'
    contour_color: str = 'C1'
    contour_linewidth: float = 0.2
    # To specify the contour levels, either provide a list of values or a integer number of contours between min and max
    contour_levels: Union[Iterable[float], int] = 10


class NMRPlot2D:
    '''
    Class to plot 2D NMR data.
    '''
    def __init__(self,
                nmr_data: NMRData2D,
                plot_settings: Optional[PlotSettings] = None):

        self.nmr_data = nmr_data
        # store the data as numpy arrays for plotting
        npeaks = len(self.nmr_data.peaks)
        self.x = np.zeros(npeaks)
        self.y = np.zeros(npeaks)
        self.sizes = np.zeros(npeaks)

        for i, peak in enumerate(self.nmr_data.peaks):
            self.x[i] = peak.x
            self.y[i] = peak.y
            self.sizes[i] = peak.correlation_strength


        # Use default plot settings if none are provided
        if plot_settings is None:
            plot_settings = PlotSettings()
        self.plot_settings = plot_settings

        self._initialize_plot_settings()

        # Set up the logger
        self.logger = logging.getLogger(__name__)
        # If not set, set number of legend elements to
        # minimum of number of peaks and 5
        if self.plot_settings.num_legend_elements is None:
            self.plot_settings.num_legend_elements = min(npeaks, DEFAULT_MAX_NUM_LEGEND_ELEMENTS)

    def _initialize_plot_settings(self):
        for key, value in self.plot_settings.__dict__.items():
            setattr(self, key, value)

    @styled_plot(nmr_base_style, nmr_2D_style)
    def plot(self, ax:Optional[Axes] = None):
        '''
        Plot the 2D NMR data.

        Parameters
        ----------
        ax : Optional[Axes], optional
            The axes to plot the data on. If not provided, a new figure and axis will be created.

        Returns
        -------
        fig : Figure
            The figure object
        ax : Axes
            The axis object
        '''

        #  Create a new figure and axis if not provided
        if ax is None:
            fig, ax = plt.subplots()
        elif isinstance(ax, Axes):
            fig = ax.get_figure()
        else:
            raise TypeError("ax must be an Axes object or None.")

        self.ax = ax
        self.fig = fig


        # --- plot lines at peak locations ---
        if self.plot_settings.show_lines:
            self._plot_axlines()

        # --- scatter plot of peaks ---
        if self.plot_settings.show_markers:
            self._plot_markers()


        # --- connectors if required ---
        if self.plot_settings.show_connectors:
            self._plot_connectors()

        # --- plot the axis labels ---
        # Use the x and y axis labels from the plot settings if provided, otherwise use the labels from the NMR data
        x_axis_label = self.plot_settings.x_axis_label if self.plot_settings.x_axis_label else self.nmr_data.x_axis_label
        y_axis_label = self.plot_settings.y_axis_label if self.plot_settings.y_axis_label else self.nmr_data.y_axis_label
        self.ax.set_xlabel(x_axis_label)
        self.ax.set_ylabel(y_axis_label)


        # other plot options
        if self.plot_settings.xlim:
            xlim = self.plot_settings.xlim
            # we handle the inverting of the axes later
            # so we just need to make sure the limits are in the right order here
            self.plot_settings.xlim = (min(xlim), max(xlim))
            self.ax.set_xlim(self.plot_settings.xlim)

        if self.plot_settings.ylim:
            ylim = self.plot_settings.ylim
            # we handle the inverting of the axes later
            # so we just need to make sure the limits are in the right order here
            self.plot_settings.ylim = (min(ylim), max(ylim))
            self.ax.set_ylim(self.plot_settings.ylim)

        # if shifts are plotted, invert the axes
        if self.nmr_data.is_shift:
            self.ax.invert_xaxis()
            self.ax.invert_yaxis()

        # --- plot the diagonal line ---
        xelem_same_as_yelem = self.nmr_data.xelement == self.nmr_data.yelement and self.nmr_data.xelement is not None
        if xelem_same_as_yelem and self.plot_settings.show_diagonal:
            # use self.xlim and self.ylim to draw a diagonal line
            ylims = self.ax.get_ylim()
            xlims = self.ax.get_xlim()
            self.ax.plot(xlims, ylims, ls='--', c='k', lw=1, alpha=0.2)

        if self.plot_settings.show_heatmap or self.plot_settings.show_contour:
            X, Y, Z = self._get_contour_data()

        # --- heatmap of peaks ---
        if self.plot_settings.show_heatmap:
            self._plot_heatmap(X, Y, Z)

        if self.plot_settings.show_contour:
            self._plot_contour(X, Y, Z)

        # --- plot the site annotations ---
        if self.plot_settings.show_labels:
            self._plot_annotations(optimise=self.plot_settings.auto_adjust_labels)


        # Display or save the plot
        self.fig.tight_layout()

        if self.plot_settings.plot_filename:
            self.fig.savefig(self.plot_settings.plot_filename, dpi=300, bbox_inches='tight')

        return self.fig, ax
    def _plot_axlines(self):
        #  we don't want to plot identical lines multiple times
        xticks = np.unique(np.round(self.x, 6))
        yticks = np.unique(np.round(self.y, 6))
        # Plot lines at the peak locations
        for x in xticks:
            self.ax.axvline(x, zorder=0)
        for y in yticks:
            self.ax.axhline(y, zorder=0)

    def _get_contour_data(self):
        xlims = self.ax.get_xlim()
        ylims = self.ax.get_ylim()
        if self.plot_settings.x_broadening is None:
            # set the broadening to 5% of the range
            self.plot_settings.x_broadening = 0.05 * abs(xlims[1] - xlims[0])
        if self.plot_settings.y_broadening is None:
            # set the broadening to 5% of the range
            self.plot_settings.y_broadening = 0.05 * abs(ylims[1] - ylims[0])

        X, Y, Z = generate_contour_map(
                    self.nmr_data.peaks,
                    grid_size = self.plot_settings.heatmap_grid_size,
                    broadening = self.plot_settings.broadening_type,
                    x_broadening=self.plot_settings.x_broadening,
                    y_broadening=self.plot_settings.y_broadening,
                    xlims = self.plot_settings.xlim,
                    ylims = self.plot_settings.ylim)
        return X, Y, Z

    def _plot_heatmap(self, X, Y, Z):
        # Plot the heatmap
        if isinstance(self.plot_settings.heatmap_levels, int):
            self.plot_settings.heatmap_levels = np.linspace(Z.min(), Z.max(), self.plot_settings.heatmap_levels)
        cs = self.ax.contourf(X, Y, Z, cmap=self.plot_settings.colormap, zorder=-1, levels=self.plot_settings.heatmap_levels)

    def _plot_contour(self, X, Y, Z):
        # fig.colorbar(cs, cax=cbar_ax, orientation='vertical', label='Intensity')
        if isinstance(self.plot_settings.contour_levels, (int, float)):
            self.plot_settings.contour_levels = np.linspace(Z.min(), Z.max(), self.plot_settings.contour_levels)

        # Add contour lines
        if self.plot_settings.show_contour:
            self.ax.contour(
                X, Y, Z,
                colors=self.plot_settings.contour_color,
                linewidths=self.plot_settings.contour_linewidth,
                levels=self.plot_settings.contour_levels
                )

    def _plot_connectors(self):
        if self.plot_settings.show_connectors and self.nmr_data.yaxis_order == '2Q' and self.nmr_data.xelement == self.nmr_data.yelement:
            x = self.x
            y = self.y
            y_order = np.argsort(y)
            # loop over peaks and plot lines between peaks with the same y value
            for i, idx in enumerate(y_order):
                if np.isclose(y[idx], y[y_order[i-1]], atol=1e-6):
                    self.ax.plot([x[idx], x[y_order[i-1]]],
                            [y[idx], y[y_order[i-1]]],
                            c='0.25',
                            lw=0.75,
                            ls='-',
                            zorder=1)

    def _plot_markers(self):
        if self.plot_settings.marker_color is None:
            # use peak colors
            color = [peak.color for peak in self.nmr_data.peaks]
            # if all colors are the same, use unique color
            if len(set(color)) == 1:
                color = color[0]

        else:
            color = self.plot_settings.marker_color

        scatter = self.ax.scatter(
                        self.x,
                        self.y,
                        s=self._normalize_marker_sizes(self.sizes),
                        c=color,
                        marker=self.plot_settings.marker,
                        linewidths=self.plot_settings.marker_linewidth,
                        zorder=10)
        if self.plot_settings.show_legend:
            # produce a legend with a cross-section of sizes from the scatter
            kw = dict(prop="sizes", num=self.plot_settings.num_legend_elements, color=color,
                      fmt=self.nmr_data.correlation_fmt + f" {self.nmr_data.correlation_unit}",
                      func=lambda s: s*np.abs(self.sizes).max() / self.plot_settings.max_marker_size)
            handles, labels = scatter.legend_elements(**kw) # type: ignore
            self.ax.legend(handles, labels,
                      title=self.nmr_data.correlation_label,
                      fancybox=True,
                      framealpha=0.8).set_zorder(12)

    def _normalize_marker_sizes(self, sizes):
        # Normalize the marker sizes, making sure all sizes are positive
        sizes = np.abs(sizes)
        marker_size_range = np.max(sizes) - np.min(sizes)
        self.logger.info(f"Marker size range: {marker_size_range} {self.nmr_data.correlation_unit}")
        max_abs_marker = np.max(sizes)
        # Normalize the marker sizes such that the maximum marker size is self.plot_settings.max_marker_size
        return sizes / max_abs_marker * self.plot_settings.max_marker_size

    def _plot_annotations(self, unique=True, optimise = True, labels_offset = 0.10):
        '''
        Get the annotations for the plot
        '''
        self.annotations = []

        # scale general font size by ANNOTATION_FONT_SCALE unless plot settings are provided
        font_size = self.plot_settings.label_fontsize
        if font_size is None:
            font_size = self.ax.xaxis.label.get_fontsize() * ANNOTATION_FONT_SCALE # type: ignore

        xpos, ypos = self.x, self.y
        xpos_label = xpos.copy()
        ypos_label = ypos.copy()

        xlabels = [peak.xlabel for peak in self.nmr_data.peaks] # type: ignore
        ylabels = [peak.ylabel for peak in self.nmr_data.peaks] # type: ignore

        # get the unique labels, keeping the indices
        if unique:
            xlabels, xidx = np.unique(xlabels, return_index=True)
            ylabels, yidx = np.unique(ylabels, return_index=True)
            # update the positions
            xpos = xpos[xidx]
            ypos = ypos[yidx]
            xpos_label = xpos_label[xidx]
            ypos_label = ypos_label[yidx]

        # we might still have some ylabels that are not the same but are at exactly the same position
        # so we need to check for this
        if len(ylabels) != len(np.unique(np.round(ypos,6))):
            # randomly perturb those y positions that are the same
            unique_ypos, idx, counts = np.unique(np.round(ypos,6), return_index=True, return_counts=True)
            for i, count in enumerate(counts):
                if count > 1:
                    ypos_label[idx[i]:idx[i]+count] += np.random.uniform(-0.5, 0.5, count)


        self.logger.debug(f'X labels: {xlabels}')
        self.logger.debug(f'X positions: {xpos}')
        self.logger.debug(f'Y labels: {ylabels}')
        self.logger.debug(f'Y positions: {ypos}')

        # TODO make a dynamical way to set the armA and armB value
        # based on the plot size
        if self.plot_settings.plot_filename is None:
            armA = 15
            armB = 15
        elif self.plot_settings.plot_filename.endswith('.pdf'):
            armA = 3
            armB = 5
        else:
            armA = 20
            armB = 30

        ######## Create x labels at the top axis ##########
        texts = []
        for i, xlabel in enumerate(xlabels):
            an = self.ax.annotate(
                xlabel,
                xy=(xpos[i], 1.0),  # Position of the annotation
                xycoords=('data', 'axes fraction'),  # Coordinate system for the annotation # type: ignore
                xytext=(xpos_label[i], 1+labels_offset),  # Position of the text
                textcoords=('data', 'axes fraction'),  # Coordinate system for the text # type: ignore
                fontsize=font_size,  # Font size of the text
                ha='center',  # Horizontal alignment
                va='bottom',  # Vertical alignment
                rotation=90,  # Rotate the text 90 degrees
                arrowprops=dict(
                    arrowstyle="-",  # Style of the arrow
                    connectionstyle=f"arc,angleA=-90,armA={armA},angleB=90,armB={armB},rad=0",  # Connection style of the arrow
                    relpos=(0.5, 0.0),  # Relative position of the arrow
                    lw=ANNOTATION_LINE_WIDTH,  # Line width of the arrow
                    shrinkA=0.0,  # Shrink factor at the start of the arrow
                    shrinkB=0.0,  # Shrink factor at the end of the arrow
                ),
            )
            texts.append(an)  # Add the annotation to the list

        if optimise:
            # Adjust the text annotations to avoid overlap
            adjust_text(
                texts,
                ensure_inside_axes=False,
                avoid_self=False,
                force_pull=(0.0, 0.0),
                force_text=(0.3, 0.0),
                force_explode=(1.5, 0.0),
                expand=(1.3, 1.0),
                max_move=2,
            )
        self.annotations += texts  # Add the adjusted texts to the annotations

        ######## Create y labels at the right axis ##########
        texts = []
        for i, ylabel in enumerate(ylabels):
            an = self.ax.annotate(
                ylabel,
                xy=(1.0, ypos[i]),  # Position of the annotation
                xycoords=('axes fraction', 'data'),  # Coordinate system for the annotation # type: ignore
                xytext=(1+labels_offset, ypos_label[i]),  # Position of the text
                textcoords=('axes fraction', 'data'),  # Coordinate system for the text # type: ignore
                fontsize=font_size,  # Font size of the text
                ha='left',  # Horizontal alignment
                va='center',  # Vertical alignment
                arrowprops=dict(
                    arrowstyle="-",  # Style of the arrow
                    connectionstyle=f"arc,angleA=180,armA={armA},angleB=0,armB={armB},rad=0",  # Connection style of the arrow
                    relpos=(0.0, 0.5),  # Relative position of the arrow
                    lw=ANNOTATION_LINE_WIDTH,  # Line width of the arrow
                    shrinkA=0.0,  # Shrink factor at the start of the arrow
                    shrinkB=0.0,  # Shrink factor at the end of the arrow
                ),
            )
            texts.append(an)  # Add the annotation to the list

        if optimise:
            # Adjust the text annotations to avoid overlap
            adjust_text(
                texts,
                ensure_inside_axes=False,
                avoid_self=False,
                force_pull=(0.0, 0.0),
                force_text=(0.4, 0.8),
                force_explode=(0.0, 1.2),
                expand=(1.0, 1.8),
                max_move=1,
            )
        self.annotations += texts  # Add the adjusted texts to the annotations
